"""Flying 200m and lap time trial module for trackmeet."""

# Refer: UCI Regulations Part 3 "Track Races" 3.2.022 - 3.2.028
# 	 and 3.2.253 - 3.2.258

import gtk
import glib
import gobject
import os
import logging

import metarace
from metarace import tod
from metarace import timy
from metarace import scbwin
from metarace import loghandler
from metarace import uiutil
from metarace import strops
from metarace import timerpane
from metarace import report
from metarace import jsonconfig

_log = logging.getLogger(u'metarace.f200')
_log.setLevel(logging.DEBUG)

# config version string
EVENT_ID = u'f200-2.0'

# startlist model columns
COL_BIB = 0
COL_FIRSTNAME = 1
COL_LASTNAME = 2
COL_CLUB = 3
COL_COMMENT = 4
COL_SEED = 5
COL_PLACE = 6
COL_START = 7
COL_FINISH = 8
COL_100 = 9

# scb function key mappings (also trig announce)
key_reannounce = u'F4'  # (+CTRL) calls into delayed announce
key_startlist = u'F3'  # startlist
key_results = u'F4'  # recalc/show result window
key_timerwin = u'F6'  # re-show timing window

# timing function key mappings
key_armstart = u'F5'  # arm for start impulse
key_armsplit = u'F7'  # de/arm intermed (manual override)
key_armfinish = u'F9'  # de/arm finish (manual override)

# extended function key mappings
key_reset = u'F5'  # + ctrl for clear/abort
key_falsestart = u'F6'  # + ctrl for false start
key_abort = u'F7'  # + ctrl abort A


class f200(object):
    """Flying 200 time trial."""

    def key_event(self, widget, event):
        """Race window key press handler."""
        if event.type == gtk.gdk.KEY_PRESS:
            key = gtk.gdk.keyval_name(event.keyval) or u'None'
            if event.state & gtk.gdk.CONTROL_MASK:
                if key == key_reset:  # override ctrl+f5
                    self.toidle()
                    return True
                elif key == key_reannounce:
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_falsestart:
                    self.falsestart()
                    return True
                elif key == key_abort:
                    self.abortrider(self.fs)
                    return True
            elif key[0] == u'F':
                if key == key_armstart:
                    self.armstart()
                    return True
                elif key == key_armsplit:
                    self.armsplit(self.fs)
                    return True
                elif key == key_armfinish:
                    self.armfinish(self.fs)
                    return True
                elif key == key_startlist:
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_timerwin:
                    self.showtimerwin()
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_results:
                    self.do_places()
                    glib.idle_add(self.delayed_announce)
                    return True
        return False

    def do_places(self):
        """Show race result on scoreboard."""
        self.meet.scbwin = None
        self.timerwin = False
        fmtplaces = []
        name_w = self.meet.scb.linelen - 11
        pcount = 0
        rcount = 0
        for r in self.riders:
            rcount += 1
            if r[COL_PLACE] is not None and r[COL_PLACE] != u'':
                pcount += 1
                plstr = r[COL_PLACE].decode(u'utf-8')
                if plstr.isdigit():
                    plstr += u'.'
                first = r[COL_FIRSTNAME].decode('utf-8')
                last = r[COL_LASTNAME].decode('utf-8')
                club = r[COL_CLUB].decode('utf-8')
                bib = r[COL_BIB].decode('utf-8')
                fmtplaces.append([
                    plstr, bib,
                    strops.fitname(first, last, name_w), club[0:3]
                ])
        FMT = [(3, u'l'), (3, u'r'), u' ', (name_w, u'l'), (4, u'r')]

        evtstatus = u'Standings'
        if rcount > 0 and pcount == rcount:
            evtstatus = u'Result'

        self.meet.scbwin = scbwin.scbtable(scb=self.meet.scb,
                                           head=self.meet.racenamecat(
                                               self.event),
                                           subhead=evtstatus.upper(),
                                           coldesc=FMT,
                                           rows=fmtplaces)
        self.meet.scbwin.reset()

    def todstr(self, col, cr, model, iter, data=None):
        """Format tod into text for listview."""
        ft = model.get_value(iter, COL_FINISH)
        if ft is not None:
            sp = model.get_value(iter, COL_100)
            st = model.get_value(iter, COL_START)
            if st is None:
                st = tod.tod(0)
            mstr = (ft - st).rawtime(3)
            sstr = u''
            if sp is not None:
                sstr = u'/' + (ft - sp).rawtime(3)
            cr.set_property(u'text', mstr + sstr)
        else:
            cr.set_property(u'text', u'')

    def loadconfig(self):
        """Load race config from disk."""
        self.riders.clear()
        self.results.clear()
        self.splits.clear()

        defautoarm = True
        defdistance = 200
        defunits = u'metres'
        defchans = 4
        defchani = 5
        defchanf = 1
        if self.evtype == u'flying lap':
            # override defaults for flying lap type
            defdistance = 1
            defunits = u'laps'
            defchans = 1
            defchani = 4
            defchanf = 1

        self.seedsrc = 1  # for autospec loads, fetch seed from the rank col

        cr = jsonconfig.config({
            u'event': {
                u'startlist': '',
                u'id': EVENT_ID,
                u'start': None,
                u'lstart': None,
                u'fsbib': None,
                u'fsstat': u'idle',
                u'showinfo': True,
                u'autospec': u'',
                u'comments': [],
                u'chan_S': defchans,
                u'chan_I': defchani,
                u'chan_F': defchanf,
                u'inomnium': False,
                u'distance': defdistance,
                u'distunits': defunits,
                u'autoarm': defautoarm
            }
        })
        cr.add_section(u'event')
        cr.add_section(u'riders')
        cr.add_section(u'traces')
        if os.path.exists(self.configfile):
            try:
                with open(self.configfile, 'rb') as f:
                    cr.read(f)
            except Exception as e:
                _log.error(u'Unable to read config: %s', e)
        else:
            _log.info(u'%r not found, loading defaults', self.configfile)

        self.chan_S = strops.confopt_chan(cr.get(u'event', u'chan_S'),
                                          defchans)
        self.chan_I = strops.confopt_chan(cr.get(u'event', u'chan_I'),
                                          defchans)
        self.chan_F = strops.confopt_chan(cr.get(u'event', u'chan_F'),
                                          defchans)
        self.comments = cr.get(u'event', u'comments')
        self.autospec = cr.get(u'event', u'autospec')
        self.distance = strops.confopt_dist(cr.get(u'event', u'distance'))
        self.units = strops.confopt_distunits(cr.get(u'event', u'distunits'))
        self.autoarm = strops.confopt_bool(cr.get(u'event', u'autoarm'))
        self.update_expander_lbl_cb()
        self.info_expand.set_expanded(
            strops.confopt_bool(cr.get(u'event', u'showinfo')))
        self.inomnium = strops.confopt_bool(cr.get(u'event', u'inomnium'))
        if self.inomnium:
            self.seedsrc = 3  # read seeding from points standinds

        # re-load starters/results and traces
        self.onestart = False
        rlist = cr.get(u'event', u'startlist').split()
        for r in rlist:
            nr = [r, u'', u'', u'', u'', u'', u'', None, None, None]
            co = u''
            st = None
            ft = None
            sp = None
            if cr.has_option(u'riders', r):
                ril = cr.get(u'riders', r)
                if len(ril) >= 1:  # save comment for stimes
                    co = ril[0]
                if len(ril) >= 2:  # write heat into rec
                    nr[COL_SEED] = ril[1]
                if len(ril) >= 4:  # Start ToD and others
                    st = tod.mktod(ril[3])
                    if st is not None:  # assigned in settimes
                        self.onestart = True
                if len(ril) >= 5:  # Finish ToD
                    ft = tod.mktod(ril[4])
                if len(ril) >= 6:  # 100m ToD
                    sp = tod.mktod(ril[5])
            dbr = self.meet.rdb.getrider(r, self.series)
            if dbr is not None:
                for i in range(1, 4):
                    nr[i] = self.meet.rdb.getvalue(dbr, i)  # unicode
            nri = self.riders.append(nr)
            self.settimes(nri, st, ft, sp, doplaces=False, comment=co)
            if cr.has_option(u'traces', r):
                self.traces[r] = cr.get(u'traces', r)
        self.placexfer()

        # re-join an existing timer state
        curstart = tod.mktod(cr.get(u'event', u'start'))
        lstart = tod.mktod(cr.get(u'event', u'lstart'))
        if lstart is None:
            lstart = curstart  # can still be None if start not set
        dorejoin = False
        fsstat = cr.get(u'event', u'fsstat')
        if fsstat not in [u'idle', u'finish']:
            self.fs.setrider(cr.get(u'event', u'fsbib'))  # load rider
            if fsstat in [u'running', u'armfin', u'armint'
                          ] and curstart is not None:
                self.fs.start(curstart)  # overrides to 'running'
                dorejoin = True

        if not self.onestart and self.autospec:
            self.meet.autostart_riders(self,
                                       self.autospec,
                                       infocol=self.seedsrc)
        if dorejoin:
            self.torunning(curstart, lstart)
        elif self.timerstat == u'idle':
            glib.idle_add(self.fs.grab_focus)

        # After load complete - check config and report.
        eid = cr.get(u'event', u'id')
        if eid and eid != EVENT_ID:
            _log.info(u'Event config mismatch: %r != %r', eid, EVENT_ID)

    def saveconfig(self):
        """Save race to disk."""
        if self.readonly:
            _log.error(u'Attempt to save readonly event')
            return
        cw = jsonconfig.config()
        cw.add_section(u'event')

        # save basic race properties
        cw.set(u'event', u'distance', self.distance)
        cw.set(u'event', u'distunits', self.units)
        cw.set(u'event', u'chan_S', self.chan_S)
        cw.set(u'event', u'chan_I', self.chan_I)
        cw.set(u'event', u'chan_F', self.chan_F)
        cw.set(u'event', u'autospec', self.autospec)
        cw.set(u'event', u'autoarm', self.autoarm)
        cw.set(u'event', u'startlist', self.get_startlist())
        cw.set(u'event', u'inomnium', self.inomnium)
        cw.set(u'event', u'showinfo', self.info_expand.get_expanded())
        cw.set(u'event', u'comments', self.comments)

        # extract and save timerpane config for interrupted run
        if self.curstart is not None:
            cw.set(u'event', u'start', self.curstart.rawtime())
        if self.lstart is not None:
            cw.set(u'event', u'lstart', self.lstart.rawtime())
        cw.set(u'event', u'fsstat', self.fs.getstatus())
        cw.set(u'event', u'fsbib', self.fs.getrider())

        cw.add_section(u'traces')
        for rider in self.traces:
            cw.set(u'traces', rider, self.traces[rider])

        cw.add_section(u'riders')

        # save out all starters
        for r in self.riders:
            # place is saved for info only
            slice = [
                r[COL_COMMENT].decode(u'utf-8'), r[COL_SEED].decode(u'utf-8'),
                r[COL_PLACE].decode(u'utf-8')
            ]
            tl = [r[COL_START], r[COL_FINISH], r[COL_100]]
            for t in tl:
                if t is not None:
                    slice.append(t.rawtime())
                else:
                    slice.append(None)
            cw.set(u'riders', r[COL_BIB].decode(u'utf-8'), slice)
        cw.set(u'event', u'id', EVENT_ID)
        _log.debug(u'Saving event config %r', self.configfile)
        with metarace.savefile(self.configfile) as f:
            cw.write(f)

    def sort_startlist(self, x, y):
        """Comparison function for seeding."""
        if x[1] == y[1]:  # same seed? revert to bib ascending
            return cmp(x[2], y[2])
        else:
            return cmp(x[1], y[1])

    def sort_heats(self, x, y):
        """Comparison function for heats."""
        (xh, xl) = strops.heatsplit(x[0])
        (yh, yl) = strops.heatsplit(y[0])
        if xh == yh:
            return cmp(xl, yl)
        else:
            return cmp(xh, yh)

    def reorder_startlist(self):
        """Re-order model according to the seeding field."""
        if len(self.riders) > 1:
            auxmap = []
            cnt = 0
            for r in self.riders:
                auxmap.append([
                    cnt,
                    strops.riderno_key(r[COL_SEED].decode(u'utf-8')),
                    strops.riderno_key(r[COL_BIB].decode(u'utf-8'))
                ])
                cnt += 1
            auxmap.sort(self.sort_startlist)
            self.riders.reorder([a[0] for a in auxmap])

    def get_heats(self, placeholders=0):
        """Return a list of heats in the event."""
        ret = []

        # arrange riders by seeding
        self.reorder_startlist()

        # then build aux map of heats
        hlist = []
        count = len(self.riders)
        if count < placeholders:
            count = placeholders
        if placeholders == 0:
            for r in self.riders:
                rno = r[COL_BIB].decode(u'utf-8')
                rh = self.meet.newgetrider(rno, self.series)
                info = None
                rname = u''
                if rh is not None:
                    rname = rh[u'namestr']
                    if rh[u'cat'] and u'tandem' in rh[u'cat'].lower():
                        ph = self.meet.newgetrider(rh[u'note'], self.series)
                        if ph is not None:
                            info = [[
                                u' ', ph[u'namestr'] + u' - Pilot',
                                ph[u'ucicode']
                            ]]
                hlist.append([str(count) + '.1', rno, rname, info])
                # all f200 heats are one up
                count -= 1
        else:
            for r in range(0, placeholders):
                rno = u''
                rname = u''
                hlist.append([str(count) + '.1', rno, rname, None])
                count -= 1

        # sort the heatlist
        hlist.sort(self.sort_heats)

        lh = None
        lcnt = 0
        rec = []
        for r in hlist:
            (h, l) = strops.heatsplit(r[0])
            if lh is not None and (h != lh or lcnt > 1):
                lcnt = 0
                ret.append(rec)
                rec = []
            heat = str(h)
            rec.extend([heat, r[1], r[2], r[3]])
            lcnt += 1
            lh = h
        if len(rec) > 0:
            ret.append(rec)
        return ret

    def startlist_report(self, program=False):
        """Return a startlist report."""
        ret = []
        cnt = 0
        sec = report.dual_ittt_startlist()
        sec.set_single()  # 200s are one-up

        headvec = [
            u'Event', self.evno, u':', self.event[u'pref'], self.event[u'info']
        ]
        if not program:
            headvec.append(u'- Start List')
        sec.heading = u' '.join(headvec)
        lapstring = strops.lapstring(self.event[u'laps'])
        substr = u' '.join(
            [lapstring, self.event[u'dist'], self.event[u'prog']]).strip()
        if substr:
            sec.subheading = substr
        if self.event[u'reco']:
            sec.footer = self.event[u'reco']
        sec.lines = self.get_heats()
        ret.append(sec)
        return ret

    def get_startlist(self):
        """Return a list of bibs in the rider model."""
        ret = []
        for r in self.riders:
            ret.append(r[COL_BIB].decode(u'utf-8'))
        return u' '.join(ret)

    def delayed_announce(self):
        """Initialise the announcer's screen after a delay."""
        if self.winopen:
            # clear page
            self.meet.txt_clear()
            self.meet.txt_title(u' '.join([
                u'Event', self.evno, u':', self.event[u'pref'],
                self.event[u'info']
            ]))
            self.meet.txt_line(1)
            self.meet.txt_line(7)

            # fill in front straight (only one?)
            fbib = self.fs.getrider()
            if fbib:
                r = self.getrider(fbib)
                if r is not None:
                    clubstr = u''
                    tcs = r[COL_CLUB].decode(u'utf-8')
                    if tcs and tcs <= 3:
                        clubstr = u'(' + tcs + u')'
                    namestr = strops.fitname(r[COL_FIRSTNAME].decode(u'utf-8'),
                                             r[COL_LASTNAME].decode(u'utf-8'),
                                             24,
                                             trunc=True)
                    placestr = u'   '  # 3 ch
                    if r[COL_PLACE]:
                        placestr = strops.truncpad(
                            r[COL_PLACE].decode(u'utf-8') + u'.', 3)
                    bibstr = strops.truncpad(r[COL_BIB].decode(u'utf-8'), 3,
                                             u'r')
                    tmstr = u''
                    et = None
                    if r[COL_START] is not None and r[COL_FINISH] is not None:
                        et = (r[COL_FINISH] - r[COL_START]).truncate(3)
                        tmstr = u'200m: ' + et.rawtime(3).rjust(12)
                    cmtstr = u''
                    if et is not None:
                        cmtstr = strops.truncpad(
                            u'Average: ' + et.speedstr(200), 38, u'r')
                    elif r[COL_COMMENT]:
                        cmtstr = strops.truncpad(
                            u'[' + r[COL_COMMENT].decode(u'utf-8').strip() +
                            u']', 38, u'r')
                    self.meet.txt_postxt(3, 0, u'        Current Rider')
                    self.meet.txt_postxt(
                        4, 0, u' '.join([placestr, bibstr, namestr, clubstr]))
                    self.meet.txt_postxt(5, 0,
                                         strops.truncpad(tmstr, 38, u'r'))
                    self.meet.txt_postxt(6, 0, cmtstr)

            # fill in leaderboard/startlist
            count = 0
            curline = 9
            posoft = 0
            for r in self.riders:
                count += 1
                if count == 19:
                    curline = 9
                    posoft += 42

                clubstr = u''
                tcs = r[COL_CLUB].decode(u'utf-8')
                if tcs and tcs <= 3:
                    clubstr = u' (' + tcs + u')'

                namestr = strops.truncpad(
                    strops.fitname(r[COL_FIRSTNAME].decode(u'utf-8'),
                                   r[COL_LASTNAME].decode(u'utf-8'),
                                   22 - len(clubstr),
                                   trunc=True) + clubstr, 22)
                placestr = u'   '  # 3 ch
                if r[COL_PLACE]:
                    placestr = strops.truncpad(
                        r[COL_PLACE].decode(u'utf-8') + u'.', 3)
                bibstr = strops.truncpad(r[COL_BIB].decode(u'utf-8'), 3, u'r')
                tmstr = u'       '  # 7 ch
                if r[COL_START] is not None and r[COL_FINISH] is not None:
                    tmstr = strops.truncpad(
                        (r[COL_FINISH] - r[COL_START]).rawtime(3), 7, u'r')
                self.meet.txt_postxt(
                    curline, posoft,
                    u' '.join([placestr, bibstr, namestr, tmstr]))
                curline += 1

    def shutdown(self, win=None, msg=u'Exiting'):
        """Terminate event object."""
        rstr = u''
        if self.readonly:
            rstr = u'readonly '
        _log.debug(u'Shutdown %sevent %s: %s', rstr, self.evno, msg)
        if not self.readonly:
            self.saveconfig()
        self.winopen = False

    def do_properties(self):
        """Run event properties dialog."""
        prfile = os.path.join(metarace.UI_PATH, u'ittt_properties.ui')
        _log.debug(u'Building event properties from %r', prfile)
        b = gtk.Builder()
        b.add_from_file(prfile)
        dlg = b.get_object(u'properties')
        dlg.set_transient_for(self.meet.window)

        # customise dialog for flying 200/flying lap
        b.get_object(u'race_score_type').hide()
        b.get_object(u'race_timing_label').hide()
        intlbltxt = u'100m Channel:'
        inthint = u'Select timing channel for 100m split.'
        if self.evtype == u'flying lap':
            intlbltxt = u'200m Channel:'
            inthint = u'Select timing channel for 200m start.'
        intlbl = b.get_object(u'race_achan_label')
        intlbl.set_text(intlbltxt)
        intcombo = b.get_object(u'race_achan_combo')
        intcombo.set_property(u'tooltip_text', inthint)
        intcombo.set_active(self.chan_I)
        finlbl = b.get_object(u'race_bchan_label')
        finlbl.set_text(u'Finish Channel:')
        fincombo = b.get_object(u'race_bchan_combo')
        fincombo.set_property(u'tooltip_text',
                              u'Select timing channel for finish.')
        fincombo.set_active(self.chan_F)
        stcombo = b.get_object(u'race_stchan_combo')
        stcombo.set_active(self.chan_S)
        aa = b.get_object(u'race_autoarm_toggle')
        aa.set_active(self.autoarm)

        di = b.get_object(u'race_dist_entry')
        if self.distance is not None:
            di.set_text(str(self.distance))
        else:
            di.set_text(u'')
        du = b.get_object(u'race_dist_type')
        if self.units == u'laps':
            du.set_active(1)
        else:
            du.set_active(0)
        se = b.get_object(u'race_series_entry')
        se.set_text(self.series)
        as_e = b.get_object(u'auto_starters_entry')
        as_e.set_text(self.autospec)

        response = dlg.run()
        if response == 1:  # id 1 set in glade for "Apply"
            dval = di.get_text().decode(u'utf-8')
            if dval.isdigit():
                self.distance = int(dval)
            else:
                self.distance = None
            if du.get_active() == 0:
                self.units = u'metres'
            else:
                self.units = u'laps'
            self.autoarm = aa.get_active()
            self.chan_S = stcombo.get_active()
            self.chan_I = intcombo.get_active()
            self.chan_F = fincombo.get_active()

            # update series
            ns = se.get_text().decode(u'utf-8')
            if ns != self.series:
                self.series = ns
                self.event[u'seri'] = ns

            # update auto startlist spec
            nspec = as_e.get_text().decode(u'utf-8')
            if nspec != self.autospec:
                self.autospec = nspec
                if self.autospec:
                    self.meet.autostart_riders(self,
                                               self.autospec,
                                               infocol=self.seedsrc)

            # xfer starters if not empty
            slist = strops.riderlist_split(
                b.get_object(u'race_starters_entry').get_text().decode(
                    u'utf-8'), self.meet.rdb, self.series)

            # if no starters yet - automatically seed by order entered
            if len(self.riders) == 0:
                cnt = 1
                for s in slist:
                    self.addrider(s, cnt)
                    cnt += 1
            else:
                for s in slist:
                    self.addrider(s)
            glib.idle_add(self.delayed_announce)
        else:
            _log.debug(u'Edit event properties cancelled')

        # if prefix is empty, grab input focus
        if not self.prefix_ent.get_text():
            self.prefix_ent.grab_focus()
        dlg.destroy()

    def result_gen(self):
        """Generator function to export rankings."""
        for r in self.riders:
            bib = r[COL_BIB].decode(u'utf-8')
            rank = None
            time = None
            info = None
            cmts = r[COL_COMMENT].decode(u'utf-8')
            if cmts in [u'rel']:
                info = cmts
            if self.onestart:
                pls = r[COL_PLACE].decode(u'utf-8')
                if pls:
                    if pls.isdigit():
                        rank = int(pls)
                    else:
                        rank = pls
                if r[COL_FINISH] is not None:
                    time = (r[COL_FINISH] - r[COL_START]).truncate(3)

            yield [bib, rank, time, info]

    def result_report(self, recurse=False):
        """Return a list of report sections containing the race result."""
        self.placexfer()
        ret = []
        sec = report.section()
        sec.heading = u'Event ' + self.evno + u': ' + u' '.join(
            [self.event[u'pref'], self.event[u'info']]).strip()
        lapstring = strops.lapstring(self.event[u'laps'])
        substr = u' '.join(
            [lapstring, self.event[u'dist'], self.event[u'prog']]).strip()
        sec.lines = []
        ftime = None
        rcount = 0
        pcount = 0
        for r in self.riders:
            rcount += 1
            rno = r[COL_BIB].decode(u'utf-8')
            rh = self.meet.newgetrider(rno, self.series)
            if rh is None:
                _log.warning(u'Rider not found %r', rno)
                continue

            rcat = None  # should check ev[u'cate']
            plink = u''
            rank = None

            rname = rh[u'namestr']
            # consider partners here
            if rh[u'cat'] and u'tandem' in rh[u'cat'].lower():
                ph = self.meet.newgetrider(rh[u'note'], self.series)
                if ph is not None:
                    plink = [
                        u'', u'', ph[u'namestr'] + u' - Pilot', ph[u'ucicode'],
                        u'', u'', u''
                    ]
            if self.event[u'cate']:
                if rh[u'cat']:
                    rcat = rh[u'cat']
            if rh[u'ucicode']:
                rcat = rh[u'ucicode']  # overwrite by force
            rtime = None
            info = None
            dtime = None
            if self.onestart:
                pls = r[COL_PLACE].decode(u'utf-8')
                if pls:
                    if pls.isdigit():
                        rank = pls + u'.'
                    else:
                        rank = pls
                    pcount += 1
                if r[COL_FINISH] is not None:
                    time = (r[COL_FINISH] - r[COL_START]).truncate(3)
                    if ftime is None:
                        ftime = time
                    else:
                        dtime = u'+' + (time - ftime).rawtime(2)
                    if r[COL_START] != tod.ZERO:
                        rtime = time.rawtime(3)
                    else:
                        rtime = time.rawtime(2) + u'\u2007'

            sec.lines.append([rank, rno, rname, rcat, rtime, dtime, plink])
        sv = []
        if substr:
            sv.append(substr)
        if self.onestart:
            if rcount > 0 and pcount < rcount:
                sv.append(u'STANDINGS')
            else:
                sv.append(u'Result')
        sec.subheading = u' - '.join(sv)

        ret.append(sec)
        return ret

    def editent_cb(self, entry, col):
        """Shared event entry update callback."""
        if col == u'pref':
            self.event[u'pref'] = entry.get_text().decode(u'utf-8')
        elif col == u'info':
            self.event[u'info'] = entry.get_text().decode(u'utf-8')
        self.update_expander_lbl_cb()

    def update_expander_lbl_cb(self):
        """Update event info expander label."""
        self.info_expand.set_label(self.meet.infoline(self.event))

    def clear_rank(self, cb):
        """Run callback once in main loop idle handler."""
        cb(u'')
        return False

    def split_trig(self, sp, t):
        """Register lap trigger."""
        bib = sp.getrider()
        self.splits.insert(t - self.curstart, None, bib)
        rank = self.splits.rank(bib)
        self.log_split(sp.getrider(), self.curstart, t)
        sp.intermed(t, 2)  # show split... delay ~2 sec
        if self.timerwin and type(self.meet.scbwin) is scbwin.scbtt:
            intstr = u' at 100m'
            if self.evtype == u'flying lap':
                intstr = u' at 200 start'
            lapstr = strops.rank2ord(unicode(rank + 1)) + intstr
            self.meet.scbwin.setr1(u'(' + unicode(rank + 1) + u')')
            glib.timeout_add_seconds(2, self.clear_rank,
                                     self.meet.scbwin.setr1)
            # announce lap and rank to txt announcer
            self.meet.txt_postxt(
                5, 8,
                strops.truncpad(lapstr, 17) + u' ' + sp.get_time())
        if self.autoarm:
            self.armfinish(sp)

    def fin_trig(self, sp, t):
        """Register finish trigger."""
        sp.finish(t)
        ri = self.getiter(sp.getrider())
        split = sp.getsplit(0)
        if ri is not None:
            self.settimes(ri, self.curstart, t, split)
        else:
            _log.warning(u'Rider %r not in model, finish time not stored',
                         sp.getrider())
        self.log_elapsed(sp.getrider(), self.curstart, t, split)
        if self.timerwin and type(self.meet.scbwin) is scbwin.scbtt:
            place = self.riders.get_value(ri, COL_PLACE)
            self.meet.scbwin.setr1(u'(' + place + u')')
            self.meet.scbwin.sett1(sp.get_time())
            dist = self.meet.get_distance(self.distance, self.units)
            if dist is not None:
                elap = t - self.curstart
                spstr = elap.speedstr(dist).strip()
                glib.timeout_add_seconds(1, self.clear_200_ttb,
                                         self.meet.scbwin, u'Avg:',
                                         spstr.rjust(12))
            else:
                glib.timeout_add_seconds(2, self.clear_200_ttb,
                                         self.meet.scbwin)
            self.meet.gemini.set_rank(place)
            self.meet.gemini.set_time((t - self.curstart).rawtime(2))
            self.meet.gemini.show_brt()

        # call for a delayed announce...
        glib.idle_add(self.delayed_announce)
        self.meet.delayed_export()

    def timercb(self, e):
        """Handle a timer event."""
        chan = timy.chan2id(e.chan)
        if self.timerstat == u'armstart':
            if chan == self.chan_S:  # Start trig
                self.torunning(e)
                glib.timeout_add_seconds(2, self.armfinish, self.fs, True)
        elif self.timerstat == u'running':
            if chan == self.chan_I:  # Intermediate
                stat = self.fs.getstatus()
                if stat == u'armint':
                    self.split_trig(self.fs, e)
                # else ignore spurious intermediate
            elif chan == self.chan_F:  # Finish
                stat = self.fs.getstatus()
                if stat in [u'armfin', u'armint']:
                    self.fin_trig(self.fs, e)
        return False

    def timeout(self):
        """Update scoreboard and respond to timing events."""
        if not self.winopen:
            return False
        if self.fs.status in [u'running', u'armint', u'armfin']:
            now = tod.now()
            self.fs.runtime(now - self.lstart)
            if self.timerwin and type(self.meet.scbwin) is scbwin.scbtt:
                self.meet.scbwin.sett1(self.fs.get_time())
                elapstr = (now - self.lstart).rawtime(1).rjust(4) + ' '
                self.meet.gemini.set_time(elapstr)
                self.meet.gemini.show_brt()
        return True

    def show_200_ttb(self, scb):
        """Display time to beat."""
        if len(self.results) > 0:
            scb.setr2(u'Fastest:')
            scb.sett2(self.results[0][0].timestr(3))
        return False

    def clear_200_ttb(self, scb, r2=u'', t2=u''):
        """Clear time to beat."""
        scb.setr2(r2)
        scb.sett2(t2)
        return False

    def torunning(self, st, lst=None):
        """Set timer running."""
        if self.fs.status == u'armstart':
            self.fs.start(st)
        self.curstart = st
        if lst is None:
            lst = tod.now()
        self.lstart = lst
        self.timerstat = u'running'
        self.onestart = True
        if self.autoarm:
            self.armsplit(self.fs)
        if self.timerwin and type(self.meet.scbwin) is scbwin.scbtt:
            glib.timeout_add_seconds(3, self.show_200_ttb, self.meet.scbwin)

    def clearplaces(self):
        """Clear rider places."""
        for r in self.riders:
            r[COL_PLACE] = u''

    def getrider(self, bib):
        """Return temporary reference to model row."""
        ret = None
        for r in self.riders:
            if r[COL_BIB].decode(u'utf-8') == bib:
                ret = r
                break
        return ret

    def addrider(self, bib=u'', info=None):
        """Add specified rider to race model."""
        istr = u''
        if info is not None:
            istr = unicode(info)
        nr = [bib, u'', u'', u'', u'', istr, u'', None, None, None]
        ri = self.getrider(bib)
        if ri is None:  # adding a new record
            dbr = self.meet.rdb.getrider(bib, self.series)
            if dbr is not None:
                for i in range(1, 4):
                    nr[i] = self.meet.rdb.getvalue(dbr, i)
            self.riders.append(nr)
        else:
            # rider exists in model, just update the seed value
            ri[COL_SEED] = istr

    def editcol_db(self, cell, path, new_text, col):
        """Cell update with writeback to meet."""
        new_text = new_text.decode(u'utf-8').strip()
        self.riders[path][col] = new_text
        rno = self.riders[path][COL_BIB].decode(u'utf-8')
        glib.idle_add(self.meet.rider_edit, rno, self.series, col, new_text)

    def placexfer(self):
        """Transfer places into model."""
        self.clearplaces()
        count = 0
        place = 1
        for t in self.results:
            bib = t[0].refid
            if t[0] > tod.FAKETIMES[u'max']:
                if t[0] == tod.FAKETIMES[u'dsq']:
                    place = u'dsq'
                elif t[0] == tod.FAKETIMES[u'ntr']:
                    place = u'ntr'
                elif t[0] == tod.FAKETIMES[u'rel']:
                    place = u'rel'
                elif t[0] == tod.FAKETIMES[u'dns']:
                    place = u'dns'
                elif t[0] == tod.FAKETIMES[u'dnf']:
                    place = u'dnf'
            else:
                place = self.results.rank(bib) + 1
            i = self.getiter(bib)
            if i is not None:
                if place == u'comment':  # superfluous but ok
                    place = self.riders.get_value(i, COL_COMMENT)
                self.riders.set_value(i, COL_PLACE, unicode(place))
                self.riders.swap(self.riders.get_iter(count), i)
                count += 1
            else:
                _log.warning(u'Rider %r not found in model, check places', bib)

    def getiter(self, bib):
        """Return temporary iterator to model row."""
        i = self.riders.get_iter_first()
        while i is not None:
            if self.riders.get_value(i, COL_BIB).decode(u'utf-8') == bib:
                break
            i = self.riders.iter_next(i)
        return i

    def settimes(self,
                 iter,
                 st=None,
                 ft=None,
                 split=None,
                 doplaces=True,
                 comment=None):
        """Transfer race times into rider model."""
        bib = self.riders.get_value(iter, COL_BIB).decode(u'utf-8')
        # clear result for this bib
        self.results.remove(bib)
        self.splits.remove(bib)
        # assign tods
        self.riders.set_value(iter, COL_START, st)
        self.riders.set_value(iter, COL_FINISH, ft)
        self.riders.set_value(iter, COL_100, split)
        # save result
        if st is None:
            st = tod.ZERO
        if ft is not None:
            last100 = None
            if split is not None:
                self.splits.insert(split - st, None,
                                   bib)  # save first 100 split
                last100 = ft - split  # and prepare to record second 100
            self.results.insert(ft - st, last100, bib)
        else:  # DNF/etc
            self.results.insert(comment, None, bib)
        # copy annotation into model if provided, or clear
        if comment:
            self.riders.set_value(iter, COL_COMMENT, comment)
        else:
            self.riders.set_value(iter, COL_COMMENT, u'')
        # if reqd, do places
        if doplaces:
            self.placexfer()

    def armstart(self):
        """Arm timer for start trigger."""
        if self.timerstat == u'armstart':
            self.toload()
        elif self.timerstat in [u'load', u'idle']:
            self.toarmstart()

    def armsplit(self, sp):
        """Arm timer for a split."""
        if self.timerstat == u'running':
            if sp.getstatus() == u'running':
                sp.toarmint(u'100m Armed')
                self.meet.main_timer.arm(self.chan_I)
            elif sp.getstatus() == u'armint':
                sp.torunning()
                self.meet.main_timer.dearm(self.chan_I)
                self.meet.main_timer.dearm(self.chan_F)
        return False

    def abortrider(self, sp):
        """Abort the current heat."""
        if sp.getstatus() not in [u'idle', u'finish']:
            bib = sp.getrider()
            ri = self.getiter(bib)
            if ri is not None:
                self.settimes(ri, st=self.curstart, comment=u'dnf')
            sp.tofinish()
            self.meet.timer_log_msg(bib, u'- Abort -')
            glib.idle_add(self.delayed_announce)

    def falsestart(self):
        """Register false start."""
        if self.timerstat == u'running':
            if self.fs.getstatus() not in [u'idle', u'finish']:
                self.fs.toload()
                bib = fs.getrider()
                self.meet.timer_log_msg(self.fs.getrider(), u'- False start -')
                if self.timerwin and type(self.meet.scbwin) is scbwin.scbtt:
                    self.meet.scbwin.setr1(u'False')
                    self.meet.scbwin.sett1(u'Start')
            self.toidle(idletimers=False)
        elif self.timerstat == u'armstart':
            if self.timerwin and type(self.meet.scbwin) is scbwin.scbtt:
                self.meet.scbwin.sett1(u'            ')
                self.meet.scbwin.sett2(u'            ')
            self.toload()

    def armfinish(self, sp, force=False):
        """Arm timer for finish trigger."""
        if self.timerstat == u'running':
            if sp.getstatus() in [u'running', u'finish']:
                if sp.getstatus() == u'finish':
                    self.meet.timer_log_msg(sp.getrider(), u'- False Finish -')
                    self.meet.scbwin.setr1(u'')
                    self.meet.scbwin.setr2(u'')
                sp.toarmfin()
                self.meet.main_timer.arm(self.chan_F)
            elif sp.getstatus() == u'armfin' and not force:
                sp.torunning()
                self.meet.main_timer.dearm(self.chan_F)
            else:
                # request to arm finish before intermediate
                self.meet.main_timer.arm(self.chan_F)
        return False

    def toload(self):
        """Set timer status to load."""
        if self.fs.status == u'armstart':
            self.fs.toload()
        self.toidle(idletimers=False)

    def fmtridername(self, tp):
        """Prepare rider name for display on scoreboard."""
        name_w = self.meet.scb.linelen - 9
        bib = tp.getrider().strip()
        if bib != u'':
            name = u'[New Rider]'
            r = self.getrider(bib)
            if r is not None and r[COL_BIB]:
                first = r[COL_FIRSTNAME].decode(u'utf-8')
                last = r[COL_LASTNAME].decode(u'utf-8')
                club = r[COL_CLUB].decode(u'utf-8')
                name = strops.fitname(first, last, name_w)
            return u' '.join([
                strops.truncpad(r[COL_BIB].decode(u'utf-8'), 3, u'r'),
                strops.truncpad(name, name_w),
                strops.truncpad(club[0:3], 4, u'r')
            ])
        else:
            return u''

    def showtimerwin(self):
        """Show timer window on scoreboard."""
        self.meet.scbwin = None
        self.meet.scbwin = scbwin.scbtt(self.meet.scb,
                                        self.meet.racenamecat(self.event),
                                        self.fmtridername(self.fs))
        self.meet.gemini.reset_fields()
        self.meet.gemini.set_bib(self.fs.getrider())
        self.meet.gemini.show_brt()
        self.timerwin = True
        self.meet.scbwin.reset()

    def toarmstart(self):
        """Set timer to arm start."""
        if self.fs.status == u'load':
            self.meet.timer_log_event(self.event)
            self.fs.toarmstart()
            self.timerstat = u'armstart'
            self.curstart = None
            self.lstart = None
            self.meet.main_timer.arm(self.chan_S)
            self.showtimerwin()
            self.meet.delayimp(u'0.01')
            if self.fs.status == u'armstart':
                bib = self.fs.getrider()
                if bib not in self.traces:
                    self.traces[bib] = []
                self.fslog = loghandler.traceHandler(self.traces[bib])
                logging.getLogger().addHandler(self.fslog)
                self.meet.scbwin.sett1(u'       0.0     ')
                nstr = self.fs.biblbl.get_text()
                self.meet.timer_log_msg(bib, nstr)
                self.meet.gemini.set_bib(bib)
                self.meet.gemini.set_time(u' 0.0 ')
                self.meet.gemini.set_rank(u'')
                self.meet.gemini.show_brt()
            glib.idle_add(self.delayed_announce)

    def toidle(self, idletimers=True):
        """Set timer to idle state."""
        if self.fslog is not None:
            logging.getLogger().removeHandler(self.fslog)
            self.fslog = None
        if idletimers:
            self.fs.toidle()
        self.timerstat = u'idle'
        self.meet.delayimp(u'2.00')
        self.curstart = None
        self.lstart = None
        for i in range(0, 8):
            self.meet.main_timer.dearm(i)
        if not self.onestart:
            pass
        self.fs.grab_focus()

    def lanelookup(self, bib=None):
        """Prepare name string for timer lane."""
        r = self.getrider(bib)
        if r is None:
            if self.meet.get_clubmode():  # fill in starters
                _log.warning(u'Adding non-starter %r', bib)
                self.addrider(bib)
                r = self.getrider(bib)
            else:
                _log.warning(u'Rider %r not in event', bib)
                return None
        rtxt = u'[New Rider]'
        if r is not None:
            rtxt = strops.listname(r[COL_FIRSTNAME].decode(u'utf-8'),
                                   r[COL_LASTNAME].decode(u'utf-8'),
                                   r[COL_CLUB].decode(u'utf-8'))
        return rtxt

    def bibent_cb(self, entry, tp):
        """Bib entry callback."""
        bib = entry.get_text().decode(u'utf-8').strip()
        if bib and bib.isalnum():
            nstr = self.lanelookup(bib)
            if nstr is not None:
                tp.biblbl.set_text(nstr)
                if tp.status == u'idle':
                    tp.toload()
                if self.timerstat == u'running':
                    tp.start(self.curstart)
            else:
                _log.warning(u'Rider %r not in event.', bib)
                tp.toidle()
        else:
            tp.toidle()

    def time_context_menu(self, widget, event, data=None):
        """Popup menu for result list."""
        self.context_menu.popup(None, None, None, event.button, event.time,
                                selpath)

    def treeview_button_press(self, treeview, event):
        """Set callback for mouse press on model view."""
        if event.button == 3:
            pathinfo = treeview.get_path_at_pos(int(event.x), int(event.y))
            if pathinfo is not None:
                path, col, cellx, celly = pathinfo
                treeview.grab_focus()
                treeview.set_cursor(path, col, 0)
                self.context_menu.popup(None, None, None, event.button,
                                        event.time)
                return True
        return False

    def tod_context_clear_activate_cb(self, menuitem, data=None):
        """Clear times for selected rider."""
        sel = self.view.get_selection().get_selected()
        if sel is not None:
            self.settimes(sel[1])
            self.log_clear(
                self.riders.get_value(sel[1], COL_BIB).decode(u'utf-8'))
            glib.idle_add(self.delayed_announce)

    def tod_context_rel_activate_cb(self, menuitem, data=None):
        """Relegate rider."""
        sel = self.view.get_selection().get_selected()
        if sel is not None:
            self.settimes(sel[1], comment=u'rel')
            glib.idle_add(self.delayed_announce)

    def tod_context_ntr_activate_cb(self, menuitem, data=None):
        """No time recorded for rider."""
        sel = self.view.get_selection().get_selected()
        if sel is not None:
            self.settimes(sel[1], comment=u'ntr')
            glib.idle_add(self.delayed_announce)

    def tod_context_dsq_activate_cb(self, menuitem, data=None):
        """Disqualify rider."""
        sel = self.view.get_selection().get_selected()
        if sel is not None:
            self.settimes(sel[1], comment=u'dsq')
            glib.idle_add(self.delayed_announce)

    def tod_context_dns_activate_cb(self, menuitem, data=None):
        """Rider did not start."""
        sel = self.view.get_selection().get_selected()
        if sel is not None:
            self.settimes(sel[1], comment=u'dns')
            glib.idle_add(self.delayed_announce)

    def tod_context_print_activate_cb(self, menuitem, data=None):
        """Print Rider trace"""
        sel = self.view.get_selection().get_selected()
        if sel is not None:
            bib = self.riders.get_value(sel[1], COL_BIB).decode(u'utf-8')
            if bib in self.traces:
                # TODO: replace timy reprint with report
                _log.info(u'CREATE AND PRINT TRACE REPORT')

    def now_button_clicked_cb(self, button, entry=None):
        """Set specified entry to the current time."""
        if entry is not None:
            entry.set_text(tod.now().timestr())

    def tod_context_edit_activate_cb(self, menuitem, data=None):
        """Run edit time dialog."""
        sel = self.view.get_selection().get_selected()
        if sel is not None:
            i = sel[1]  # grab off row iter
            stod = self.riders.get_value(i, COL_START)
            st = u''
            if stod is not None:
                st = stod.timestr()
            ftod = self.riders.get_value(i, COL_FINISH)
            ft = u''
            if ftod is not None:
                ft = ftod.timestr()
            ret = uiutil.edit_times_dlg(self.meet.window, st, ft)
            if ret[0] == 1:
                stod = tod.mktod(ret[1])
                ftod = tod.mktod(ret[2])
                bib = self.riders.get_value(i, COL_BIB).decode(u'utf-8')
                if stod is not None and ftod is not None:
                    self.settimes(i, stod, ftod)  # set times
                    self.log_elapsed(bib, stod, ftod, manual=True)
                else:
                    self.settimes(i)  # clear times
                    self.log_clear(bib)
                _log.info(u'Race times manually adjusted for rider %r', bib)
            else:
                _log.debug('Edit race times cancelled')
            glib.idle_add(self.delayed_announce)

    def tod_context_del_activate_cb(self, menuitem, data=None):
        """Delete selected row from race model."""
        sel = self.view.get_selection().get_selected()
        if sel is not None:
            i = sel[1]  # grab off row iter
            self.riders.remove(i)
            glib.idle_add(self.delayed_announce)

    def log_clear(self, bib):
        """Print clear time log."""
        self.meet.timer_log_msg(bib, u'- Time Cleared -')

    def log_split(self, bib, start, split):
        """Print split log."""
        slbl = u'100'
        if self.evtype == u'flying lap':
            slbl = u'int'
        self.meet.timer_log_straight(bib, slbl, split - start, 3)

    def log_elapsed(self, bib, start, finish, split=None, manual=False):
        """Print elapsed log info."""
        if manual:
            self.meet.timer_log_msg(bib, u'- Manual Adjust -')
        self.meet.timer_log_straight(bib, u'ST', start)
        self.meet.timer_log_straight(bib, u'FIN', finish)
        if split is not None:
            slbl = u'L100'
            if self.evtype == u'flying lap':
                slbl = u'L200'
            self.meet.timer_log_straight(bib, slbl, finish - split, 3)
        self.meet.timer_log_straight(bib, u'TIME', finish - start, 3)

    def destroy(self):
        """Signal race shutdown."""
        if self.context_menu is not None:
            self.context_menu.destroy()
        self.frame.destroy()

    def show(self):
        """Show race window."""
        self.frame.show()

    def hide(self):
        """Hide race window."""
        self.frame.hide()

    def __init__(self, meet, event, ui=True):
        """Constructor."""
        self.meet = meet
        self.event = event  # Note: now a treerowref
        self.evno = event[u'evid']
        self.evtype = event[u'type']
        self.series = event[u'seri']
        self.configfile = meet.event_configfile(self.evno)
        self.readonly = not ui
        rstr = u''
        if self.readonly:
            rstr = u'readonly '
        _log.debug(u'Init %sevent %s', rstr, self.evno)

        # properties
        self.distance = 200
        self.units = u'metres'
        self.autoarm = True
        self.comments = []
        self.chan_S = 4
        self.chan_I = 5
        self.chan_F = 1

        # race run time attributes
        self.onestart = False
        self.winopen = ui
        self.timerwin = False
        self.timerstat = u'idle'
        self.curstart = None
        self.lstart = None
        self.results = tod.todlist(u'FIN')
        self.splits = tod.todlist(u'100')
        self.autospec = u''
        self.inomnium = False
        self.seedsrc = 1  # default seeding is by rank in last round

        self.riders = gtk.ListStore(
            gobject.TYPE_STRING,  # 0 bib
            gobject.TYPE_STRING,  # 1 firstname
            gobject.TYPE_STRING,  # 2 lastname
            gobject.TYPE_STRING,  # 3 club
            gobject.TYPE_STRING,  # 4 Comment
            gobject.TYPE_STRING,  # 5 seed
            gobject.TYPE_STRING,  # 6 place
            gobject.TYPE_PYOBJECT,  # 7 Start
            gobject.TYPE_PYOBJECT,  # 8 Finish
            gobject.TYPE_PYOBJECT)  # 9 100m

        uifile = os.path.join(metarace.UI_PATH, u'ittt.ui')
        _log.debug(u'Building event interface from %r', uifile)
        b = gtk.Builder()
        b.add_from_file(uifile)

        self.frame = b.get_object(u'race_vbox')
        self.frame.connect(u'destroy', self.shutdown)

        # meta info pane
        self.info_expand = b.get_object(u'info_expand')
        b.get_object(u'race_info_evno').set_text(self.evno)
        self.showev = b.get_object(u'race_info_evno_show')
        self.prefix_ent = b.get_object(u'race_info_prefix')
        self.prefix_ent.connect(u'changed', self.editent_cb, u'pref')
        self.prefix_ent.set_text(self.event[u'pref'])
        self.info_ent = b.get_object(u'race_info_title')
        self.info_ent.connect(u'changed', self.editent_cb, u'info')
        self.info_ent.set_text(self.event[u'info'])
        self.traces = {}

        # Timer Pane
        mf = b.get_object(u'race_timer_pane')
        self.fs = timerpane.timerpane(u'Timer')
        self.fs.bibent.connect(u'activate', self.bibent_cb, self.fs)
        self.fs.hide_splits()
        self.fs.splitlbls = [u'100\u2006m Split', u'Finish']
        self.fslog = None
        mf.pack_start(self.fs.frame)

        # show window
        self.context_menu = None
        if ui:
            _log.debug(u'Connecting event ui handlers')
            # Result Pane
            t = gtk.TreeView(self.riders)
            self.view = t
            t.set_reorderable(True)
            t.set_rules_hint(True)
            t.connect(u'button_press_event', self.treeview_button_press)

            # TODO: show team name & club but pop up for rider list
            tmlbl = u'200m/Last 100m'
            if self.evtype == u'flying lap':
                tmlbl = u'Time/Last 200m'
            uiutil.mkviewcoltxt(t, u'No.', COL_BIB, calign=1.0)
            uiutil.mkviewcoltxt(t,
                                u'First Name',
                                COL_FIRSTNAME,
                                self.editcol_db,
                                expand=True)
            uiutil.mkviewcoltxt(t,
                                u'Last Name',
                                COL_LASTNAME,
                                self.editcol_db,
                                expand=True)
            uiutil.mkviewcoltxt(t, u'Club', COL_CLUB, self.editcol_db)
            uiutil.mkviewcoltxt(t, u'Seed', COL_SEED, self.editcol_db)
            uiutil.mkviewcoltod(t, tmlbl, cb=self.todstr)
            uiutil.mkviewcoltxt(t, u'Rank', COL_PLACE, halign=0.5, calign=0.5)
            t.show()
            b.get_object(u'race_result_win').add(t)

            b.connect_signals(self)
            b = gtk.Builder()
            b.add_from_file(os.path.join(metarace.UI_PATH, u'tod_context.ui'))
            self.context_menu = b.get_object(u'tod_context')
            b.connect_signals(self)



"""Shared gtk UI helper functions."""

import os
import gtk
import gobject
from pango import (FontDescription, STYLE_OBLIQUE, STYLE_NORMAL)
import metarace
from metarace import tod
from metarace import strops

# Font-overrides
DIGITFONT = FontDescription(u'Noto Mono Medium 22')
MONOFONT = FontDescription(u'Noto Mono')
LOGVIEWFONT = FontDescription(u'Noto Mono 11')

# Button indications
bg_none = gtk.image_new_from_file(metarace.default_file(u'bg_idle.svg'))
bg_armstart = gtk.image_new_from_file(
    metarace.default_file(u'bg_armstart.svg'))
bg_armint = gtk.image_new_from_file(metarace.default_file(u'bg_armint.svg'))
bg_armfin = gtk.image_new_from_file(metarace.default_file(u'bg_armfin.svg'))


def hvscroller(child):
    """Return a new scrolled window packed with the supplied child."""
    vs = gtk.ScrolledWindow()
    vs.show()
    vs.set_border_width(5)
    vs.set_shadow_type(gtk.SHADOW_IN)
    vs.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
    vs.add(child)
    return vs


def vscroller(child):
    """Return a new scrolled window packed with the supplied child."""
    vs = gtk.ScrolledWindow()
    vs.show()
    vs.set_border_width(5)
    vs.set_shadow_type(gtk.SHADOW_IN)
    vs.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
    vs.add(child)
    return vs


class statbut(object):

    def __init__(self, b, w=12):
        c = gtk.HBox(spacing=2)
        self._i = gtk.image_new_from_pixbuf(bg_none.get_pixbuf())
        #self._i = gtk.image_new_from_file(metarace.default_file(u'bg_idle.svg'))
        self._i.show()
        c.pack_start(self._i, False, True, 0)
        self._l = gtk.Label(u'Idle')
        self._l.set_width_chars(w)
        self._l.set_single_line_mode(True)
        self._l.show()
        c.pack_start(self._l, True, True, 0)
        c.show()
        b.add(c)
        self._b = b

    def buttonchg(self, image, label=None):
        self._i.set_from_pixbuf(image.get_pixbuf())
        if label is not None:
            self._l.set_text(label)

    def set_sensitive(self, sensitive=False):
        self._b.set_sensitive(sensitive)


def mkviewcoltod(view=None,
                 header=u'',
                 cb=None,
                 width=120,
                 editcb=None,
                 colno=None):
    """Return a Time of Day view column."""
    i = gtk.CellRendererText()
    i.set_property(u'xalign', 1.0)
    j = gtk.TreeViewColumn(header, i)
    j.set_cell_data_func(i, cb, colno)
    if editcb is not None:
        i.set_property(u'editable', True)
        i.connect(u'edited', editcb, colno)
    j.set_min_width(width)
    view.append_column(j)
    return j


def mkviewcoltxt(view=None,
                 header=u'',
                 colno=None,
                 cb=None,
                 width=None,
                 halign=None,
                 calign=None,
                 expand=False,
                 editcb=None,
                 maxwidth=None,
                 bgcol=None,
                 fontdesc=None,
                 fixed=False):
    """Return a text view column."""
    i = gtk.CellRendererText()
    if cb is not None:
        i.set_property(u'editable', True)
        i.connect(u'edited', cb, colno)
    if calign is not None:
        i.set_property(u'xalign', calign)
    if fontdesc is not None:
        i.set_property(u'font_desc', fontdesc)
    j = gtk.TreeViewColumn(header, i, text=colno)
    if bgcol is not None:
        j.add_attribute(i, u'background', bgcol)
    if halign is not None:
        j.set_alignment(halign)
    if fixed:
        j.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
    if expand:
        if width is not None:
            j.set_min_width(width)
        j.set_expand(True)
    else:
        if width is not None:
            j.set_min_width(width)
    if maxwidth is not None:
        j.set_max_width(maxwidth)
    view.append_column(j)
    if editcb is not None:
        i.connect(u'editing-started', editcb)
    return i


def mkviewcolbg(view=None,
                header=u'',
                colno=None,
                cb=None,
                width=None,
                halign=None,
                calign=None,
                expand=False,
                editcb=None,
                maxwidth=None):
    """Return a text view column."""
    i = gtk.CellRendererText()
    if cb is not None:
        i.set_property(u'editable', True)
        i.connect(u'edited', cb, colno)
    if calign is not None:
        i.set_property(u'xalign', calign)
    j = gtk.TreeViewColumn(header, i, background=colno)
    if halign is not None:
        j.set_alignment(halign)
    if expand:
        if width is not None:
            j.set_min_width(width)
        j.set_expand(True)
    else:
        if width is not None:
            j.set_min_width(width)
    if maxwidth is not None:
        j.set_max_width(maxwidth)
    view.append_column(j)
    if editcb is not None:
        i.connect(u'editing-started', editcb)
    return i


def savecsvdlg(title=u'', parent=None, hintfile=None, lpath=None):
    ret = None
    dlg = gtk.FileChooserDialog(title, parent, gtk.FILE_CHOOSER_ACTION_SAVE,
                                (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                 gtk.STOCK_SAVE, gtk.RESPONSE_OK))
    cfilt = gtk.FileFilter()
    cfilt.set_name(u'CSV Files')
    cfilt.add_mime_type(u'text/csv')
    cfilt.add_pattern(u'*.csv')
    dlg.add_filter(cfilt)
    cfilt = gtk.FileFilter()
    cfilt.set_name(u'All Files')
    cfilt.add_pattern(u'*')
    dlg.add_filter(cfilt)
    if lpath:
        dlg.set_current_folder(lpath)
    if hintfile:
        dlg.set_current_name(hintfile)
    response = dlg.run()
    if response == gtk.RESPONSE_OK:
        ret = dlg.get_filename().decode(u'utf-8')
    dlg.destroy()
    return ret


def loadcsvdlg(title=u'', parent=None, lpath=None):
    ret = None
    dlg = gtk.FileChooserDialog(title, parent, gtk.FILE_CHOOSER_ACTION_OPEN,
                                (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                 gtk.STOCK_OPEN, gtk.RESPONSE_OK))
    cfilt = gtk.FileFilter()
    cfilt.set_name(u'CSV Files')
    cfilt.add_mime_type(u'text/csv')
    cfilt.add_pattern(u'*.csv')
    dlg.add_filter(cfilt)
    cfilt = gtk.FileFilter()
    cfilt.set_name(u'All Files')
    cfilt.add_pattern(u'*')
    dlg.add_filter(cfilt)
    if lpath:
        dlg.set_current_folder(lpath)
    response = dlg.run()
    if response == gtk.RESPONSE_OK:
        ret = dlg.get_filename().decode(u'utf-8')
    dlg.destroy()
    return ret


def mkviewcolbool(view=None,
                  header=u'',
                  colno=None,
                  cb=None,
                  width=None,
                  expand=False):
    """Return a boolean view column."""
    i = gtk.CellRendererToggle()
    i.set_property(u'activatable', True)
    if cb is not None:
        i.connect(u'toggled', cb, colno)
    j = gtk.TreeViewColumn(header, i, active=colno)
    if expand:
        j.set_min_width(width)
        j.set_expand(True)
    else:
        if width is not None:
            j.set_min_width(width)
    view.append_column(j)
    return i


def coltxtbibser(col, cr, model, iter, data):
    """Display a bib.ser string in a tree view."""
    (bibcol, sercol) = data
    cr.set_property(
        u'text',
        strops.bibser2bibstr(model.get_value(iter, bibcol),
                             model.get_value(iter, sercol)))


def mkviewcolbibser(view=None,
                    header=u'No.',
                    bibno=0,
                    serno=1,
                    width=None,
                    expand=False):
    """Return a column to display bib/series as a bib.ser string."""
    i = gtk.CellRendererText()
    i.set_property(u'xalign', 1.0)
    j = gtk.TreeViewColumn(header, i)
    j.set_cell_data_func(i, coltxtbibser, (bibno, serno))
    if expand:
        j.set_min_width(width)
        j.set_expand(True)
    else:
        if width is not None:
            j.set_min_width(width)
    view.append_column(j)
    return i


def mktextentry(prompt, row, table):
    """Create and return a text entry within a gtk table."""
    if u'?' not in prompt:
        prompt += u':'
    l = gtk.Label(prompt)
    l.set_alignment(0.0, 0.5)
    l.show()
    table.attach(l, 0, 1, row, row + 1, gtk.FILL, gtk.FILL, xpadding=5)
    e = gtk.Entry()
    e.set_width_chars(24)
    e.set_activates_default(True)  # Check assumption on window
    e.show()
    table.attach(e,
                 1,
                 2,
                 row,
                 row + 1,
                 gtk.FILL | gtk.EXPAND,
                 gtk.FILL,
                 xpadding=5,
                 ypadding=2)
    return e


def mkcomboentry(prompt, row, table, options):
    """Create and return a combo entry within a gtk table."""
    l = gtk.Label(prompt)
    l.set_alignment(1.0, 0.5)
    l.show()
    table.attach(l, 0, 1, row, row + 1, gtk.FILL, gtk.FILL, xpadding=5)
    c = gtk.combo_box_new_text()
    c.show()
    for opt in options:
        c.append_text(opt)
    table.attach(c, 1, 2, row, row + 1, gtk.FILL, gtk.FILL, xpadding=5)
    return c


def mklbl(prompt, row, table):
    """Create and return label within a gtk table."""
    l = gtk.Label(prompt)
    l.set_alignment(1.0, 0.5)
    l.show()
    table.attach(l, 0, 1, row, row + 1, gtk.FILL, gtk.FILL, xpadding=5)
    e = gtk.Label()
    e.set_alignment(0.0, 0.5)
    e.show()
    table.attach(e, 1, 2, row, row + 1, gtk.FILL, gtk.FILL, xpadding=5)
    return e


def mkbutintbl(prompt, row, col, table):
    """Create and return button within a gtk table."""
    b = gtk.Button(prompt)
    b.show()
    table.attach(b,
                 col,
                 col + 1,
                 row,
                 row + 1,
                 gtk.FILL,
                 gtk.FILL,
                 xpadding=5,
                 ypadding=5)
    return b


def questiondlg(window, question, subtext=None):
    """Display a question dialog and return True/False."""
    dlg = gtk.MessageDialog(window, gtk.DIALOG_MODAL, gtk.MESSAGE_QUESTION,
                            gtk.BUTTONS_YES_NO, question)
    if subtext is not None:
        dlg.format_secondary_text(subtext)
    ret = False
    response = dlg.run()
    if response == gtk.RESPONSE_YES:
        ret = True
    dlg.destroy()
    return ret


def now_button_clicked_cb(button, entry=None):
    """Copy the current time of day into the supplied entry."""
    if entry is not None:
        entry.set_text(tod.now().timestr())


def edit_times_dlg(window,
                   stxt=None,
                   ftxt=None,
                   btxt=None,
                   ptxt=None,
                   bonus=False,
                   penalty=False,
                   finish=True):
    """Display times edit dialog and return updated time strings."""
    b = gtk.Builder()
    b.add_from_file(os.path.join(metarace.UI_PATH, u'edit_times.ui'))
    dlg = b.get_object(u'timing')
    dlg.set_transient_for(window)

    se = b.get_object(u'timing_start_entry')
    se.modify_font(MONOFONT)
    if stxt is not None:
        se.set_text(stxt)
    b.get_object(u'timing_start_now').connect(u'clicked',
                                              now_button_clicked_cb, se)

    fe = b.get_object(u'timing_finish_entry')
    fe.modify_font(MONOFONT)
    if ftxt is not None:
        fe.set_text(ftxt)
    if finish:
        b.get_object(u'timing_finish_now').connect(u'clicked',
                                                   now_button_clicked_cb, fe)
    else:
        b.get_object(u'timing_finish_now').set_sensitive(False)

    be = b.get_object(u'timing_bonus_entry')
    be.modify_font(MONOFONT)
    if btxt is not None:
        be.set_text(btxt)
    if bonus:
        be.show()
        b.get_object(u'timing_bonus_label').show()

    pe = b.get_object(u'timing_penalty_entry')
    pe.modify_font(MONOFONT)
    if ptxt is not None:
        pe.set_text(ptxt)
    if penalty:
        pe.show()
        b.get_object(u'timing_penalty_label').show()

    ret = dlg.run()
    stxt = se.get_text().decode(u'utf-8').strip()
    ftxt = fe.get_text().decode(u'utf-8').strip()
    btxt = be.get_text().decode(u'utf-8').strip()
    ptxt = pe.get_text().decode(u'utf-8').strip()
    dlg.destroy()
    return (ret, stxt, ftxt, btxt, ptxt)



"""Meet loader application."""

import pygtk

pygtk.require("2.0")

import gtk
import glib
import gobject

import os
import sys
import logging
import datetime
import metarace

from metarace import jsonconfig
from metarace import strops
from metarace import uiutil
from metarace import eventdb
from metarace import trackmeet
from metarace import roadmeet

_log = logging.getLogger(u'metarace.apploader')
_log.setLevel(logging.DEBUG)


def meet_identify(configpath=None):
    """Try to identify the meet located at configpath, returning info."""
    # Returns an array:
    # [type, configpath, descr, evtinfo, datestr]
    ret = None
    cfile = os.path.join(configpath, u'config.json')
    efile = os.path.join(configpath, u'events.csv')
    if os.path.exists(cfile):
        cr = jsonconfig.config()
        try:
            _log.debug(u'Reading meet config %r', cfile)
            with open(cfile, 'rb') as f:
                cr.read(f)
            edb = eventdb.eventdb()
            _log.debug(u'Reading meet events %r', efile)
            edb.load(efile)
            if cr.has_section(u'trackmeet'):  # this is TRACK
                descr = u''
                if cr.has_option(u'trackmeet', u'title'):
                    descr = cr.get(u'trackmeet', u'title') + u' '
                if cr.has_option(u'trackmeet', u'subtitle'):
                    descr += cr.get(u'trackmeet', u'subtitle') + u' '
                ecstr = u'New Meet'
                ecount = len(edb)
                if ecount:
                    ecstr = unicode(ecount) + u' events'
                infostr = u''
                if cr.has_option(u'trackmeet', u'date'):
                    infostr = cr.get(u'trackmeet', u'date')
                _log.debug(u'Track meet: %s/%s/%s', descr, ecstr, infostr)
                ret = [u'track', configpath, descr.strip(), ecstr, infostr]
            elif cr.has_section(u'roadmeet'):  # this is Road
                etype = u'road'
                eh = edb.getfirst()
                if eh is not None and eh[u'type']:
                    etype = eh[u'type']
                descr = u''
                if cr.has_option(u'roadmeet', u'title'):
                    descr = cr.get(u'roadmeet', u'title') + u' '
                if cr.has_option(u'roadmeet', u'subtitle'):
                    descr += cr.get(u'roadmeet', u'subtitle') + u' '
                infostr = u''
                if cr.has_option(u'roadmeet', u'date'):
                    infostr = cr.get(u'roadmeet', u'date')
                _log.debug(u'Road meet: %s/%s/%s', descr, etype, infostr)
                ret = [u'road', configpath, descr.strip(), etype, infostr]
            else:
                _log.warning(u'Meet type could not be identified')
        except Exception as e:
            _log.error(u'Error reading meet folder %r: %s', configpath, e)
    return ret


class apploader(object):
    """Meet loader."""

    def destroy_cb(self, window, msg=u''):
        """Handle destroy signal and exit application."""
        self.window.hide()
        glib.idle_add(self.destroy_handler)

    def create_new_meet(self, path=None):
        """Create new meet of the specified type at path."""
        if path is None or not path:
            path = unicode(datetime.datetime.now().strftime('%F_%H%M%S'))

        # load up dialog and fill path
        b = gtk.Builder()
        b.add_from_file(os.path.join(metarace.UI_PATH, 'create_meet.ui'))
        dlg = b.get_object('create')
        dlg.set_transient_for(self.window)
        etsel = b.get_object('create_type')
        etsel.set_active(1)
        etmodel = b.get_object('type_model')
        pent = b.get_object('path_entry')
        pent.set_text(path)

        response = dlg.run()
        dlg.hide()
        if response == 1:
            np = pent.get_text().decode(u'utf8')
            nt = etmodel[etsel.get_active()][0]
            _log.debug(u'Create: %r / %r', nt, np)

            # if ok, create path - if necessary
            if not os.path.exists(np):
                _log.debug(u'Creating new meet path: %r', np)
                os.makedirs(np)
            oktogo = True
            # refuse to continue if path contains a config.json
            if os.path.exists(os.path.join(np, u'config.json')):
                over = uiutil.questiondlg(
                    self.window, u'Overwrite existing meet?',
                    u'The chosen path appears to already contain a meet. Overwriting the meet will destroy any existing information.'
                )
                if over:
                    os.unlink(os.path.join(np, u'config.json'))
                    os.unlink(os.path.join(np, u'events.csv'))
                else:
                    oktogo = False
            # validate type
            tv = nt.split(u':')
            etype = None
            subtype = None
            if len(tv) < 1 or len(tv) > 2:
                _log.debug(u'Invalid type')
                oktogo = False
            else:
                etype = tv[0]
                if len(tv) > 1:
                    subtype = tv[1]

            if oktogo:
                _log.info(u'Create %r:%r at: %r', etype, subtype, np)
                cw = jsonconfig.config()
                if etype == u'track':
                    cw.add_section(u'trackmeet')
                else:
                    cw.add_section(u'roadmeet')
                    if subtype:
                        edb = eventdb.eventdb([])
                        event = edb.add_empty(evno=u'00')
                        event[u'type'] = subtype
                        edb.save(os.path.join(np, u'events.csv'))
                with open(os.path.join(np, u'config.json'), 'wb') as f:
                    cw.write(f)
            glib.idle_add(self.loadpath)
        else:
            _log.debug(u'CREATE: cancelled.')

        dlg.destroy()
        return False  # if called from idle_add

    def create_but_clicked_cb(self, button, data=None):
        """Handle create new meet."""
        self.create_new_meet()

    def show(self):
        self.window.show()

    def shutdown(self, data=None):
        pass

    def destroy_handler(self):
        gtk.main_quit()
        return False

    def infostring(self, maintype, subtype):
        """Return a human-readable type string."""
        ret = u'Unknown'
        if maintype == u'road':
            if subtype == u'handicap':
                ret = u'Road Handicap'
            elif subtype == u'road':
                ret = u'Road Race'
            elif subtype == u'irtt':
                ret = u'Road Time Trial'
            elif subtype == u'trtt':
                ret = u'Team Time Trial'
            elif subtype == u'criterium':
                ret = u'Criterium'
            elif subtype == u'circuit':
                ret = u'Circuit'
            elif subtype == u'cross':
                ret = u'Cyclocross'
        elif maintype == u'track':
            ret = u'Track (' + subtype.title() + ')'
        return ret

    def run_meet(self, etype, epath, data=None):
        """Load and run the selected existing meet."""
        cfpath = os.path.abspath(epath)
        _log.debug(u'Running %r meet: %r', etype, cfpath)
        if etype == u'track':
            self.subapp = trackmeet.runapp(cfpath)
        else:
            self.subapp = roadmeet.runapp(cfpath)
        self.window.hide()

    def row_activated_cb(self, view, path, column, data=None):
        """Handle activate on a view row."""
        row = self.model[path]
        self.run_meet(row[6], row[0])

    def open_but_clicked_cb(self, button, data=None):
        """Handle click on open button."""
        # If Row Selected, try to open that one, else open file chooser..
        sel = self.view.get_selection()
        cnt = sel.count_selected_rows()
        if cnt > 0:
            (model, iters) = sel.get_selected_rows()
            elist = [(model[i][6], model[i][0]) for i in iters]
            if len(elist) == 1:
                _log.debug(u'Opening meet: %r', elist)
                self.run_meet(elist[0][0], elist[0][1])
                return True
            else:
                _log.error('Unable to open meet: %r', elist)
                return False
        else:
            _log.debug(u'No selection, continue to file open dialog')
        path = None
        dlg = gtk.FileChooserDialog(u'Open Metarace Meet', self.window,
                                    gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                    (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                     gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        dlg.set_current_folder(self.path)
        response = dlg.run()
        if response == gtk.RESPONSE_OK:
            path = dlg.get_filename()
        else:
            pass
        dlg.destroy()

        if path is not None:
            meet = meet_identify(path)
            if meet is not None:
                self.run_meet(meet[0], meet[1])
            else:
                self.create_new_meet(path)
                # one level of repeat for the special case
                meet = meet_identify(path)
                if meet is not None:
                    self.run_meet(meet[0], meet[1])

    def loadpath(self, data=None):
        """Load the meet list."""
        self.model.clear()
        for root, dirs, files in os.walk(self.path):
            if root == self.path:
                for d in dirs:
                    if d != metarace.DEFAULTS_NAME:  # ignore sysdefaults
                        meet = meet_identify(os.path.join(root, d))
                        if meet is not None:
                            tinfo = self.infostring(meet[0], meet[3])
                            self.model.append([
                                meet[1], d, meet[2], tinfo, meet[4], u'',
                                meet[0], u''
                            ])
            del dirs[0:]
        return False

    def __init__(self, path=None):
        """Apploader constructor."""
        self.path = u'.'
        if path is not None:
            self.path = path
        self.subapp = None
        b = gtk.Builder()
        b.add_from_file(os.path.join(metarace.UI_PATH, u'apploader.ui'))
        self.window = b.get_object('window')

        self.model = gtk.ListStore(
            gobject.TYPE_STRING,  # fullpath
            gobject.TYPE_STRING,  # shortpath
            gobject.TYPE_STRING,  # descr
            gobject.TYPE_STRING,  # type
            gobject.TYPE_STRING,  # date
            gobject.TYPE_STRING,  # extra info
            gobject.TYPE_STRING,  # class road/track
            gobject.TYPE_STRING)  # sorting?
        t = gtk.TreeView(self.model)
        #t.set_reorderable(True)
        t.set_rules_hint(True)
        #t.set_headers_clickable(True)
        self.model.set_sort_column_id(1, gtk.SORT_ASCENDING)
        #t.set_headers_visible(False)
        t.set_search_column(1)
        uiutil.mkviewcoltxt(t, 'Path', 1, width=100)
        uiutil.mkviewcoltxt(t, 'Description', 2, expand=True, maxwidth=300)
        uiutil.mkviewcoltxt(t, 'Type', 3, width=100)
        t.connect('row-activated', self.row_activated_cb)
        t.show()
        self.view = t
        b.get_object('scrollbox').add(t)
        b.connect_signals(self)
        glib.idle_add(self.loadpath)


def appmain():
    configpath = metarace.config_path(metarace.DATA_PATH)
    if configpath is None:
        _log.error(u'Unable to open directory %r', sys.argv[1])
        sys.exit(1)
    os.chdir(configpath)
    app = apploader()
    app.show()
    try:
        metarace.mainloop()
    except:
        app.shutdown(u'Exception from main loop.')
        if app.subapp is not None:
            app.subapp.shutdown()
        raise


def meetmain(cfpath):
    configpath = metarace.config_path(cfpath)
    if configpath is None:
        _log.error(u'Unable to open meet config %r', sys.argv[1])
        sys.exit(1)
    app = None
    meet = meet_identify(configpath)
    if meet is not None:
        _log.info(u'Loading %s', meet[2])
        if meet[0] == u'track':
            app = trackmeet.runapp(configpath)
        else:
            app = roadmeet.runapp(configpath)
    try:
        metarace.mainloop()
    except:
        app.shutdown(u'Exception from main loop')
        raise


def main():
    """Run the app loader."""
    # attach a console log handler to the root logger
    ch = logging.StreamHandler()
    ch.setLevel(metarace.LOGLEVEL)
    fh = logging.Formatter(metarace.LOGFORMAT)
    ch.setFormatter(fh)
    logging.getLogger().addHandler(ch)
    metarace.init(withgtk=True)
    # check for a provided path
    if len(sys.argv) > 2:
        print(u'Usage: metarace [configdir]\n')
        sys.exit(1)
    elif len(sys.argv) == 2:
        meetmain(sys.argv[1])
    else:
        appmain()


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: MIT
"""Telegraph

 Telegraph is a thin wrapper over paho.mqtt.client, which provides
 a MQTT pub/sub interface, altered for use with metarace applications.

 Example:

        import metarace
        from metarace import telegraph
        metarace.init()

        def messagecb(topic, message):
            obj = telegraph.from_json(message)
            ...
        t = telegraph.telegraph()
        t.set_will_json({u'example':[]}, u'thetopic')
        t.subscribe(u'thetopic')
        t.setcb(messagecb)
        t.start()
        ...
        t.publish_json({u'example':[1,2,3]}, u'thetopic')

 Message callback functions receive two named parameters 'topic' and
 'message' which are both unicode strings. The message callback is run in
 the telegraph thread context. Use the convenience function "from_json"
 to convert a message from json into a python object. See defcallback
 for an example.

 Configuration is via metarace system config (metarace.json), under
 section 'telegraph':

  key: (type) Description [default]
  --
  host : (string) MQTT broker, None to disable ['localhost']
  port: (int) MQTT port [1883/8883]
  usetls : (bool) if True, connect to server over TLS [False]
  debug : (bool) if True, enable logging in MQTT library [False]
  username : (string) username [None]
  password : (string) password [None]
  deftopic : (string) a default publish topic [None]
  persist : (bool) if true, open a persistent connection to broker [False]
  clientid : (string) provide an explicit client id [None]
  qos : (int) default QOS to use for subscribe and publish [0]


"""

import threading
import Queue
import logging
import json
import paho.mqtt.client as mqtt
from uuid import uuid4
import metarace

QUEUE_TIMEOUT = 2

# module logger
_log = logging.getLogger(u'metarace.telegraph')
_log.setLevel(logging.DEBUG)


def from_json(payload=None):
    """Return message payload decoded from json, or None."""
    ret = None
    try:
        ret = json.loads(payload)
    except Exception as e:
        _log.warning(u'%s decoding JSON payload: %s', e.__class__.__name__, e)
    return ret


def defcallback(topic=None, message=None):
    """Default message receive callback function."""
    ob = from_json(message)
    if ob is not None:
        _log.debug(u'RCV %r: %r', topic, ob)
    else:
        _log.debug(u'RCV %r: %r', topic, message)


class telegraph(threading.Thread):
    """Metarace telegraph server thread."""

    def subscribe(self, topic=None, qos=None):
        """Add topic to the set of subscriptions."""
        if topic:
            self.__subscriptions[topic] = qos
            if self.__connected:
                self.__queue.put_nowait((u'SUBSCRIBE', topic, qos))

    def unsubscribe(self, topic=None):
        """Remove topic from the set of subscriptions."""
        if topic and topic in self.__subscriptions:
            del self.__subscriptions[topic]
            if self.__connected:
                self.__queue.put_nowait(('UNSUBSCRIBE', topic))

    def setcb(self, func=None):
        """Set the message receive callback function."""
        if func is not None:
            self.__cb = func
        else:
            self.__cb = defcallback

    def set_deftopic(self, topic=None):
        """Set or clear the default publish topic."""
        if isinstance(topic, basestring) and topic:
            self.__deftopic = topic
        else:
            self.__deftopic = None
        _log.debug(u'Default publish topic set to: %r', self.__deftopic)

    def set_will_json(self, obj=None, topic=None, qos=None, retain=False):
        """Pack the provided object into JSON and set as will."""
        try:
            self.set_will(json.dumps(obj), topic, qos, retain)
        except Exception as e:
            _log.error(u'Error setting will object %r: %s', obj, e)

    def set_will(self, message=None, topic=None, qos=None, retain=False):
        """Set or clear the last will with the broker."""
        if not self.__connect_pending and not self.__connected:
            if topic is not None:
                nqos = qos
                if nqos is None:
                    nqos = self.__qos
                payload = None
                if message is not None:
                    payload = message.encode('utf-8')
                self.__client.will_set(topic, payload, nqos, retain)
                _log.debug(u'Will set on topic %r', topic)
            else:
                self.__client.will_clear()
                _log.debug(u'Cleared will')
        else:
            _log.error(u'Unable to set will, already connected')

    def connected(self):
        """Return true if connected."""
        return self.__connected

    def reconnect(self):
        """Request re-connection to broker."""
        self.__queue.put_nowait((u'RECONNECT', True))

    def exit(self, msg=None):
        """Request thread termination."""
        self.__running = False
        self.__queue.put_nowait((u'EXIT', msg))

    def wait(self):
        """Suspend calling thread until command queue is processed."""
        self.__queue.join()

    def publish(self, message=None, topic=None, qos=None, retain=False):
        """Publish the provided msg to topic or deftopic if None."""
        self.__queue.put_nowait((u'PUBLISH', topic, message, qos, retain))

    def publish_json(self, obj=None, topic=None, qos=None, retain=False):
        """Pack the provided object into JSON and publish to topic."""
        try:
            self.publish(unicode(json.dumps(obj)), topic, qos, retain)
        except Exception as e:
            _log.error(u'Error publishing object %r: %s', obj, e)

    def __init__(self):
        """Constructor."""
        threading.Thread.__init__(self)
        self.daemon = True
        self.__queue = Queue.Queue()
        self.__cb = defcallback
        self.__subscriptions = {}
        self.__deftopic = None
        self.__connected = False
        self.__connect_pending = False
        self.__host = 'localhost'
        self.__port = 1883
        self.__cid = None
        self.__persist = False
        self.__resub = True
        self.__qos = 0
        self.__doreconnect = False

        # check system config for overrides
        if metarace.sysconf.has_option(u'telegraph', u'host'):
            self.__host = metarace.sysconf.get_str(u'telegraph', u'host')
        if metarace.sysconf.has_option(u'telegraph', u'deftopic'):
            # note: this may be overidden by application
            self.__deftopic = metarace.sysconf.get_str(u'telegraph',
                                                       u'deftopic')
        if metarace.sysconf.has_option(u'telegraph', u'qos'):
            self.__qos = metarace.sysconf.get_posint(u'telegraph', u'qos', 0)
            if self.__qos > 2:
                _log.info(u'Invalid QOS %r set to %r', self.__qos, 2)
                self.__qos = 2
        if metarace.sysconf.has_option(u'telegraph', u'clientid'):
            self.__cid = metarace.sysconf.get_str(u'telegraph', u'clientid')
        if not self.__cid:
            self.__cid = str(uuid4())
        _log.debug(u'Using client id: %r', self.__cid)
        if metarace.sysconf.has_option(u'telegraph', u'persist'):
            self.__persist = metarace.sysconf.get_bool(u'telegraph',
                                                       u'persist')
        _log.debug(u'Persistent connection: %r', self.__persist)

        # create mqtt client
        self.__client = mqtt.Client(client_id=self.__cid,
                                    clean_session=not self.__persist)
        if metarace.sysconf.has_option(u'telegraph', u'debug'):
            if metarace.sysconf.get_bool(u'telegraph', u'debug'):
                _log.debug(u'Enabling mqtt/paho debug')
                mqlog = logging.getLogger(u'metarace.telegraph.mqtt')
                mqlog.setLevel(logging.DEBUG)
                self.__client.enable_logger(mqlog)
        if metarace.sysconf.has_option(u'telegraph', u'usetls'):
            if metarace.sysconf.get_bool('telegraph', 'usetls'):
                _log.debug(u'Enabling TLS connection')
                self.__port = 8883
                self.__client.tls_set()
        username = None
        password = None
        if metarace.sysconf.has_option(u'telegraph', u'username'):
            username = metarace.sysconf.get_str(u'telegraph', u'username')
        if metarace.sysconf.has_option(u'telegraph', u'password'):
            password = metarace.sysconf.get_str(u'telegraph', u'password')
        if username and password:
            self.__client.username_pw_set(username, password)
        # override automatic port selection if provided
        if metarace.sysconf.has_option(u'telegraph', u'port'):
            np = metarace.sysconf.get_posint(u'telegraph', u'port')
            if np is not None:
                self.__port = np
                _log.debug(u'Set port to %r', self.__port)
        self.__client.reconnect_delay_set(2, 16)
        self.__client.on_message = self.__on_message
        self.__client.on_connect = self.__on_connect
        self.__client.on_disconnect = self.__on_disconnect
        if self.__host:
            self.__doreconnect = True
        self.__running = False

    def __reconnect(self):
        if not self.__connect_pending:
            if self.__connected:
                _log.debug(u'Disconnecting client')
                self.__client.disconnect()
                self.__client.loop_stop()
            if self.__host:
                _log.debug(u'Connecting to %s:%d', self.__host, self.__port)
                self.__connect_pending = True
                self.__client.connect_async(self.__host, self.__port)
                self.__client.loop_start()

    # PAHO methods
    def __on_connect(self, client, userdata, flags, rc):
        if rc == 0:
            _log.debug(u'Connect %r: %r/%r', client._client_id, flags, rc)
            if not self.__resub and self.__persist and flags[
                    u'session present']:
                _log.debug(u'Resumed existing session for %r',
                           client._client_id)
            else:
                _log.debug(u'Assuming Clean session for %r', client._client_id)
                s = []
                for t in self.__subscriptions:
                    nqos = self.__subscriptions[t]
                    if nqos is None:
                        nqos = self.__qos
                    s.append((t, nqos))
                if len(s) > 0:
                    _log.debug(u'Subscribe topics: %r', s)
                    self.__client.subscribe(s)
                self.__resub = False
            self.__connected = True
        else:
            _log.info(u'Connect failed with error %r: %r', rc,
                      mqtt.connack_string(rc))
            self.__connected = False
        self.__connect_pending = False

    def __on_disconnect(self, client, userdata, rc):
        _log.debug(u'Disconnect %r: %r', client._client_id, rc)
        self.__connected = False
        # Note: PAHO lib will attempt re-connection automatically

    def __on_message(self, client, userdata, message):
        #_log.debug(u'Message from %r: %r', client._client_id, message)
        self.__cb(topic=message.topic,
                  message=message.payload.decode(u'utf-8'))

    def run(self):
        """Called via threading.Thread.start()."""
        self.__running = True
        if self.__host:
            _log.debug(u'Starting')
        else:
            _log.debug(u'Not connected')
        while self.__running:
            try:
                # Check connection status
                if self.__host and self.__doreconnect:
                    self.__doreconnect = False
                    if not self.__connect_pending:
                        self.__reconnect()
                # Process command queue
                while self.__running:
                    m = self.__queue.get(timeout=QUEUE_TIMEOUT)
                    self.__queue.task_done()
                    if m[0] == u'PUBLISH':
                        ntopic = self.__deftopic
                        if m[1] is not None:  # topic is set
                            ntopic = m[1]
                        nqos = m[3]
                        if nqos is None:
                            nqos = self.__qos
                        if ntopic:
                            msg = None
                            if m[2] is not None:
                                msg = m[2].encode(u'utf-8')
                            self.__client.publish(ntopic, msg, nqos, m[4])
                        else:
                            #_log.debug(u'No topic, msg ignored: %r', m[1])
                            pass
                    elif m[0] == u'SUBSCRIBE':
                        _log.debug(u'Subscribe topic: %r', m[1])
                        nqos = m[2]
                        if nqos is None:
                            nqos = self.__qos
                        self.__client.subscribe(m[1], nqos)
                    elif m[0] == u'UNSUBSCRIBE':
                        _log.debug(u'Un-subscribe topic: %r', m[1])
                        self.__client.unsubscribe(m[1])
                    elif m[0] == u'RECONNECT':
                        self.__connect_pending = False
                        self.__doreconnect = True
                    elif m[0] == u'EXIT':
                        _log.debug(u'Request to close: %r', m[1])
                        self.__running = False
            except Queue.Empty:
                pass
            except Exception as e:
                _log.error(u'%s: %s', e.__class__.__name__, e)
                self.__connect_pending = False
                self.__doreconnect = False
        self.__client.disconnect()
        self.__client.loop_stop()
        _log.info(u'Exiting')



# SPDX-License-Identifier: MIT
"""Transponder decoder interface."""

import threading
import Queue
import logging

from metarace import tod
from metarace import sysconf

_log = logging.getLogger(u'metarace.decoder')
_log.setLevel(logging.DEBUG)
DECODER_LOG_LEVEL = 15
logging.addLevelName(DECODER_LOG_LEVEL, u'DECODER')
_PHOTOTHRESH = tod.tod(u'0.03')
_DEFAULT_HANDLER = u'null'


def mkdevice(portstr=u'', curdev=None):
    """Return a decoder handle for the provided port specification."""
    # Note: If possible, returns the current device
    ret = curdev
    devtype = _DEFAULT_HANDLER
    if sysconf.has_option(u'decoder', u'default'):
        devtype = sysconf.get(u'decoder', u'default')
        _log.debug(u'Default type set to %r from sysconf', devtype)
    (a, b, c) = portstr.partition(u':')
    if b:
        a = a.lower()
        if a in HANDLERS:
            devtype = a
        a = c  # shift port into a
    devport = a
    if curdev is None:
        curdev = HANDLERS[devtype]()
        curdev.setport(devport)
    elif type(curdev) is HANDLERS[devtype]:
        _log.debug(u'Requested decoder is %s', curdev.__class__.__name__)
        curdev.setport(devport)
    else:
        _log.debug(u'Changing decoder type from %s to %s',
                   curdev.__class__.__name__, devtype)
        curdev.setcb(None)
        wasalive = curdev.running()
        if wasalive:
            curdev.exit(u'Change decoder type')
        curdev = None
        curdev = HANDLERS[devtype]()
        curdev.setport(devport)
        _log.debug(u'Starting %s decoder', curdev.__class__.__name__)
        if wasalive:
            curdev.start()
    return curdev


class decoder(threading.Thread):
    """Idealised transponder decoder interface."""

    # API Methods
    def getcb(self):
        """Return the current callback function."""
        return self._cb

    def setcb(self, func=None):
        """Set or clear the event callback."""
        if func is not None:
            self._cb = func
        else:
            self._cb = self._defcallback

    def running(self):
        """Return state of running flag."""
        return self._running

    def connected(self):
        """Return true if decoder instance is connected."""
        return False

    def exit(self, msg=None):
        """Request thread termination."""
        self._running = False
        self._cqueue.put_nowait((u'_exit', msg))

    def setport(self, device=None):
        """Request new device address."""
        self._flush()
        self._cqueue.put_nowait((u'_port', device))

    def sane(self, data=None):
        """Reset decoder to sane state."""
        self._cqueue.put_nowait((u'_sane', data))

    def sync(self, data=None):
        """Synchronise decoder to host PC time."""
        self._cqueue.put_nowait((u'_sync', data))

    def start_session(self, data=None):
        """Request decoder start timing session."""
        self._cqueue.put_nowait((u'_start_session', data))

    def stop_session(self, data=None):
        """Request decoder stop timing session."""
        self._cqueue.put_nowait((u'_stop_session', data))

    def status(self, data=None):
        """Request status message from decoder."""
        self._cqueue.put_nowait((u'_status', data))

    def clear(self, data=None):
        """Clear passings in decoder memory."""
        self._cqueue.put_nowait((u'_clear', data))

    def trig(self, impulse=None):
        """Queue a fake timing impulse through decoder interface."""
        self._cqueue.put_nowait((u'_trig', impulse))

    def replay(self, file=None):
        """Request replay of passings from the provided file indicator."""
        self._cqueue.put_nowait((u'_replay', file))

    def wait(self):
        """Suspend calling thread until the command queue is empty."""
        self._cqueue.join()

    def write(self, msg=None):
        """Queue a raw device command string."""
        self._cqueue.put_nowait((u'_write', msg))

    def photothresh(self):
        """Return the photo finish threshold."""
        return _PHOTOTHRESH

    # Private Methods
    def __init__(self):
        threading.Thread.__init__(self)
        self.daemon = True
        self._cqueue = Queue.Queue()
        self._running = False
        self._cb = self._defcallback

    def _defcallback(self, evt=None):
        """Default callback is a debug log entry."""
        _log.debug(unicode(evt))

    def _close(self):
        """Close hardware connection to decoder."""
        raise NotImplementedError(u'decoder._close()')

    def _flush(self):
        """Clear out the command queue."""
        try:
            while True:
                self._cqueue.get_nowait()
                self._cqueue.task_done()
        except Queue.Empty:
            pass

    def _exit(self, msg):
        """Handle request to exit."""
        _log.debug(u'Request to exit: %r', msg)
        self._close()
        self._flush()
        self._running = False

    def _port(self, port):
        """Re-connect decoder hardware."""
        raise NotImplementedError(u'decoder._port()')

    def _sane(self, data=None):
        """Return decoder to known initial state."""
        raise NotImplementedError(u'decoder._sane()')

    def _sync(self, data=None):
        """Roughly align decoder timebase with host PC clock."""
        raise NotImplementedError(u'decoder._sync()')

    def _start_session(self, data=None):
        """Start decoder timing session."""
        raise NotImplementedError(u'decoder._start_session()')

    def _stop_session(self, data=None):
        """Stop decoder timing session."""
        raise NotImplementedError(u'decoder._stop_session()')

    def _status(self, data=None):
        """Request status from decoder."""
        raise NotImplementedError(u'decoder._status()')

    def _clear(self, data=None):
        """Request clear of memory on decoder."""
        raise NotImplementedError(u'decoder._clear()')

    def _trig(self, impulse):
        """Return a timing impulse to the host application."""
        self._cb(impulse)

    def _replay(self, file):
        """Request a replay of passings from file."""
        raise NotImplementedError(u'decoder._replay()')

    def _write(self, msg):
        """Write the supplied msg to the decoder."""
        raise NotImplementedError(u'decoder._write()')

    def _proccmd(self, cmd):
        """Process a command tuple from the queue."""
        method = getattr(self, cmd[0], None)
        if method is not None:
            method(cmd[1])
        else:
            _log.debug(u'Unknown command: %r', cmd)

    def run(self):
        """Decoder main loop."""
        _log.debug(u'Starting')
        self._running = True
        while self._running:
            try:
                c = self._cqueue.get()
                self._cqueue.task_done()
                self._proccmd(c)
            except Exception as e:
                # errors from dummy decoder should not appear in UI
                _log.debug(u'%s: %s', e.__class__.__name__, e)
        self.setcb()  # make sure callback is unrefed
        _log.debug(u'Exiting')


from rrs import rrs
from rru import rru
from thbc import thbc

HANDLERS = {
    u'null': decoder,
    u'thbc': thbc,
    u'rrs': rrs,
    u'rru': rru,
}



#!/usr/bin/python2

import logging
from time import sleep
import metarace
from metarace import timy

logging.basicConfig(level=logging.DEBUG)
metarace.init()


def timercb(impulse):
    print(repr(impulse))


t = timy.timy()
t.setport(u'/dev/ttyS0')
t.setcb(timercb)
t.start()
t.sane()
t.delaytime(u'0.1')
try:
    while True:
        sleep(2)
        t.arm(u'C0')
finally:
    t.exit()
    t.join()



"""Sprint round handler for trackmeet."""

# Note: A bug in the model/interface causes incorrect results to be
#       displayed on scb if a result is changed. Work required to fix.

import gtk
import glib
import gobject
import pango
import logging
import os

import metarace
from metarace import tod
from metarace import jsonconfig
from metarace import timy
from metarace import unt4
from metarace import riderdb
from metarace import scbwin
from metarace import uiutil
from metarace import strops
from metarace import report

# config version string
EVENT_ID = 'tracksprnd-1.0'

# race gobject model column constants
COL_CONTEST = 0  # contest ID '1v16'
COL_A_NO = 1  # Number of A rider
COL_A_STR = 2  # Namestr of A rider
COL_A_PLACE = 3  # Place string of A rider
COL_B_NO = 4  # Number of B rider
COL_B_STR = 5  # Namestr of B rider
COL_B_PLACE = 6  # Place string of B rider
COL_200M = 7  # time for last 200m
COL_WINNER = 8  # no of 'winner'
COL_COMMENT = 9  # decisions of commissaire's panel

# scb function key mappings
key_startlist = 'F3'  # show starters in table
key_results = 'F4'  # recalc/show result window

# timing function key mappings
key_armstart = 'F5'  # arm for start/200m impulse
key_showtimer = 'F6'  # show timer
key_armfinish = 'F9'  # arm for finish impulse
key_win_a = 'F11'  # A rider wins
key_win_b = 'F12'  # B rider wins

# extended function key mappings
key_abort = 'F5'  # + ctrl for clear/abort
key_walk_a = 'F9'  # + ctrl for walk over
key_walk_b = 'F10'
key_rel_a = 'F11'  # + ctrl for relegation
key_rel_b = 'F12'


class sprnd(object):
    """Data handling for sprint rounds."""

    def addrider(self, bib='', info=None):
        """Add specified rider to race model."""
        rstr = u''
        rcom = u''
        cstack = []
        for cr in self.contests:
            # check for 'first' empty slot in A riders
            if cr[COL_A_NO] == u'':
                cr[COL_A_NO] = bib
                cr[COL_A_STR] = self.rider_name(bib)
                cr[COL_A_PLACE] = u''  # LOAD?
                ## special case the bye here
                if cr[COL_CONTEST] == u'bye':
                    cr[COL_A_PLACE] = u' '
                    cr[COL_B_STR] = u' '
                    cr[COL_B_PLACE] = u' '
                    cr[COL_B_NO] = u' '
                    cr[COL_WINNER] = bib  # auto win the bye rider
                return
            cstack.insert(0, cr)  # do these get re-used?
        for cr in cstack:
            # reverse contests for B riders
            if cr[COL_B_NO] == u'':
                cr[COL_B_NO] = bib
                cr[COL_B_STR] = self.rider_name(bib)
                cr[COL_B_PLACE] = u''  # LOAD?
                return
        self.log.warn('Not enough heats for the specified starters: ' +
                      repr(bib))

    def delrider(self, bib):
        """Remove specified rider from the model."""
        for c in self.contests:
            if c[COL_A_NO] == bib:
                c[COL_A_NO] = u''
                c[COL_A_PLACE] = u''
                c[COL_A_STR] = u''
            elif c[COL_B_NO] == bib:
                c[COL_B_NO] = u''
                c[COL_B_PLACE] = u''
                c[COL_B_STR] = u''
            if c[COL_WINNER] == bib:
                c[COL_200M] = None
                c[COL_WINNER] = u''

    def loadconfig(self):
        """Load race config from disk."""
        self.contests.clear()

        cr = jsonconfig.config({
            u'event': {
                u'id': EVENT_ID,
                u'contests': [],
                u'timerstat': None,
                u'showinfo': True,
                u'comment': u'',
                u'autospec': u''
            },
            u'contests': {}
        })
        cr.add_section(u'event')
        cr.add_section(u'contests')
        try:
            with open(self.configpath, 'rb') as f:
                cr.read(f)
        except Exception as e:
            self.log.error(u'Reading event config: ' + unicode(e))

        # event metas
        self.info_expand.set_expanded(
            strops.confopt_bool(cr.get(u'event', u'showinfo')))
        self.autospec = cr.get(u'event', u'autospec')
        self.comment = cr.get(u'event', u'comment')
        self.onestart = False

        # restore contests
        oft = 0
        curactive = -1
        for c in cr.get(u'event', u'contests'):
            if cr.has_option(u'contests', c):
                res = cr.get(u'contests', c)
                ft = tod.mktod(res[4])
                if ft or res[5]:
                    self.onestart = True  # at least one run so far
                else:
                    if curactive == -1:
                        curactive = oft
                astr = ''
                if res[0]:
                    astr = self.rider_name(res[0])
                bstr = ''
                if res[2]:
                    bstr = self.rider_name(res[2])
                nr = [
                    c, res[0], astr, res[1], res[2], bstr, res[3], ft, res[5],
                    res[6]
                ]
                self.add_contest(c, nr)
            else:
                self.add_contest(c)
            oft += 1

        if not self.onestart and self.autospec:
            self.del_riders()
            self.meet.autostart_riders(self, self.autospec)

        self.current_contest_combo.set_active(curactive)

        eid = cr.get('event', 'id')
        if eid and eid != EVENT_ID:
            self.log.error('Event configuration mismatch: ' + repr(eid) +
                           ' != ' + repr(EVENT_ID))
            #self.readonly = True

    def rider_name(self, bib):
        """Return a formated rider name string."""
        # TODO: Cache?
        first = ''
        last = ''
        club = ''
        dbr = self.meet.rdb.getrider(bib, self.series)
        if dbr is not None:
            first = self.meet.rdb.getvalue(dbr, 1)
            last = self.meet.rdb.getvalue(dbr, 2)
            club = self.meet.rdb.getvalue(dbr, 3)
        return strops.resname(first, last, club)

    def del_riders(self):
        """Remove all starters from model."""
        for c in self.contests:
            for col in [
                    COL_A_NO, COL_A_STR, COL_A_PLACE, COL_B_NO, COL_B_STR,
                    COL_B_PLACE, COL_WINNER
            ]:
                c[col] = u''
                c[COL_200M] = None

    def add_contest(self, c, cv=[]):
        #if c in self.contests:	# already in event	# requires scan tbl
        #self.log.info('Ignored duplicate contest: ' + repr(c))
        #return None
        if len(cv) == 10:
            self.contests.append(cv)
        else:
            self.contests.append([c, '', '', '', '', '', '', None, '', ''])

    def race_ctrl_action_activate_cb(self, entry, data=None):
        """Perform current action on bibs listed."""
        rlist = entry.get_text()
        acode = self.action_model.get_value(
            self.ctrl_action_combo.get_active_iter(), 1)
        if acode == 'add':
            rlist = strops.reformat_riderlist(rlist, self.meet.rdb,
                                              self.series)
            for bib in rlist.split():
                self.addrider(bib)
            entry.set_text('')
        elif acode == 'del':
            rlist = strops.reformat_riderlist(rlist, self.meet.rdb,
                                              self.series)
            for bib in rlist.split():
                self.delrider(bib)
            entry.set_text('')
        elif acode == 'com':  # add comment to contest
            self.add_comment(rlist)
        else:
            self.log.error('Ignoring invalid action.')
            return False
        glib.idle_add(self.delayed_announce)

    def startlist_report(self, program=False):
        """Return a startlist report."""
        ret = []
        if self.event[u'type'] == 'sprint final':
            sec = report.sprintfinal()
        else:
            sec = report.sprintround()
        headvec = [
            u'Event', self.evno, u':', self.event[u'pref'], self.event[u'info']
        ]
        if not program:
            headvec.append(u'- Start List')
        sec.heading = u' '.join(headvec)

        lapstring = strops.lapstring(self.event[u'laps'])
        substr = u' '.join(
            [lapstring, self.event[u'dist'], self.event[u'prog']]).strip()
        if substr:
            sec.subheading = substr

        for cr in self.contests:
            byeflag = None
            bno = cr[COL_B_NO]
            bname = cr[COL_B_STR].decode('utf-8')
            timestr = None
            if cr[COL_CONTEST] == u'bye':
                timestr = u' '
                bno = u' '
                bname = u' '
                byeflag = u' '
            if self.event[u'type'] == 'sprint final':
                ## HACK: assume no bye
                sec.lines.append([
                    cr[COL_CONTEST] + u':',
                    [
                        None, cr[COL_A_NO], cr[COL_A_STR].decode('utf-8'),
                        None, None, None, None
                    ],
                    [
                        None, cr[COL_B_NO], cr[COL_B_STR].decode('utf-8'),
                        None, None, None, None
                    ]
                ])
            else:
                sec.lines.append([
                    cr[COL_CONTEST] + u':',
                    [None, cr[COL_A_NO], cr[COL_A_STR].decode('utf-8'), None],
                    [byeflag, bno, bname, None], timestr
                ])
        ret.append(sec)
        return ret

    def saveconfig(self):
        """Save race to disk."""
        if self.readonly:
            self.log.error('Attempt to save readonly ob.')
            return
        cw = jsonconfig.config()
        cw.add_section(u'event')
        cw.add_section(u'contests')

        cw.set(u'event', u'showinfo', self.info_expand.get_expanded())
        cw.set(u'event', u'timerstat', self.timerstat)
        cw.set(u'event', u'comment', self.comment)
        cw.set(u'event', u'autospec', self.autospec)
        contestlist = []
        for c in self.contests:
            cid = c[COL_CONTEST].decode('utf-8')
            contestlist.append(cid)
            ft = None
            if c[COL_200M]:
                ft = c[COL_200M].rawtime()
            cw.set(u'contests', cid, [
                c[COL_A_NO], c[COL_A_PLACE], c[COL_B_NO], c[COL_B_PLACE], ft,
                c[COL_WINNER], c[COL_COMMENT]
            ])
        cw.set(u'event', u'contests', contestlist)
        cw.set(u'event', u'id', EVENT_ID)
        self.log.debug('Saving race config to: ' + self.configpath)
        with open(self.configpath, 'wb') as f:
            cw.write(f)

    def shutdown(self, win=None, msg='Exiting'):
        """Terminate race object."""
        self.log.debug(u'Race shutdown: ' + msg)  # only ref'd local
        if not self.readonly:
            self.saveconfig()
        self.winopen = False

    def do_properties(self):
        """Run race properties dialog."""
        b = gtk.Builder()
        b.add_from_file(os.path.join(metarace.UI_PATH, 'sprnd_properties.ui'))
        dlg = b.get_object('properties')
        dlg.set_transient_for(self.meet.window)
        ## FIX
        response = dlg.run()
        if response == 1:  # id 1 set in glade for "Apply"
            self.log.debug(u'Updating race properties.')
            glib.idle_add(self.delayed_announce)
        else:
            self.log.debug(u'Edit race properties cancelled.')

        # if prefix is empty, grab input focus
        if self.prefix_ent.get_text() == u'':
            self.prefix_ent.grab_focus()
        dlg.destroy()

    def resettimer(self):
        """Reset race timer."""
        self.finish = None
        self.start = None
        self.lstart = None
        self.curelap = None
        self.timerstat = 'idle'
        self.meet.timer.dearm(self.startchan)
        self.meet.timer.dearm(0)
        self.meet.timer.dearm(self.finchan)
        uiutil.buttonchg(self.stat_but, uiutil.bg_none, 'Idle')
        self.stat_but.set_sensitive(True)
        self.set_elapsed()

    def setrunning(self):
        """Set timer state to 'running'."""
        self.timerstat = 'running'
        uiutil.buttonchg(self.stat_but, uiutil.bg_none, 'Running')

    def setfinished(self):
        """Set timer state to 'finished'."""
        self.timerstat = 'finished'
        uiutil.buttonchg(self.stat_but, uiutil.bg_none, 'Finished')
        self.stat_but.set_sensitive(False)

    def armstart(self):
        """Toggle timer arm start state."""
        if self.timerstat == 'idle':
            self.timerstat = 'armstart'
            uiutil.buttonchg(self.stat_but, uiutil.bg_armstart, 'Arm Start')
            self.meet.timer.arm(self.startchan)
            self.meet.timer.arm(0)
        elif self.timerstat == 'armstart':
            self.timerstat = 'idle'
            self.time_lbl.set_text('')
            uiutil.buttonchg(self.stat_but, uiutil.bg_none, 'Idle')
            self.meet.timer.dearm(self.startchan)
            self.meet.timer.dearm(0)
        return False  # for use in delayed callback

    def armfinish(self):
        """Toggle timer arm finish state."""
        if self.timerstat == 'running':
            self.timerstat = 'armfinish'
            uiutil.buttonchg(self.stat_but, uiutil.bg_armfin, 'Arm Finish')
            self.meet.timer.arm(self.finchan)
        elif self.timerstat == 'armfinish':
            self.timerstat = 'running'
            uiutil.buttonchg(self.stat_but, uiutil.bg_none, 'Running')
            self.meet.timer.dearm(self.finchan)
        return False  # for use in delayed callback

    def showtimer(self):
        """Display the running time on the scoreboard."""
        if self.timerstat == 'idle':
            self.armstart()
        ## NOTE: display todo
        tp = '200m:'
        self.meet.scbwin = scbwin.scbtimer(self.meet.scb, self.event[u'pref'],
                                           self.event[u'info'], tp)
        self.timerwin = True
        self.meet.scbwin.reset()
        self.meet.gemini.reset_fields()
        if self.timerstat == 'finished':
            if self.start is not None and self.finish is not None:
                elap = self.finish - self.start
                self.meet.scbwin.settime(elap.timestr(2))
                self.meet.scbwin.setavg(elap.speedstr(200))  # fixed dist
                self.meet.gemini.set_time(elap.rawtime(2))
            self.meet.scbwin.update()
        self.meet.gemini.show_brt()

    def key_event(self, widget, event):
        """Race window key press handler."""
        if event.type == gtk.gdk.KEY_PRESS:
            key = gtk.gdk.keyval_name(event.keyval) or 'None'
            if event.state & gtk.gdk.CONTROL_MASK:
                if key == key_abort:  # override ctrl+f5
                    self.resettimer()
                    return True
                elif key == key_walk_a:
                    self.set_winner(u'A', wplace=u'w/o', lplace=u'dns')
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_walk_b:
                    self.set_winner(u'B', wplace=u'w/o', lplace=u'dns')
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_rel_a:
                    self.set_winner(u'B', wplace=u'1.', lplace=u'rel')
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_rel_b:  # rel B => A wins
                    self.set_winner(u'A', wplace='1.', lplace='rel')
                    glib.idle_add(self.delayed_announce)
                    return True
                # TODO: next/prev contest
            if key[0] == 'F':
                if key == key_armstart:
                    self.armstart()
                    return True
                elif key == key_armfinish:
                    self.armfinish()
                    return True
                elif key == key_showtimer:
                    self.showtimer()
                    return True
                elif key == key_startlist:
                    self.do_startlist()
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_results:
                    self.doscbplaces = True  # override if already clear
                    self.redo_places()
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_win_a:
                    self.set_winner(u'A')
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_win_b:
                    self.set_winner(u'B')
                    glib.idle_add(self.delayed_announce)
                    return True
        return False

    def set_winner(self, win, wplace=u'1.', lplace=u'2.'):
        i = self.current_contest_combo.get_active_iter()
        if i is not None:  # contest selected ok
            prevwin = self.contests.get_value(i, COL_WINNER)
            cid = self.contests.get_value(i, COL_CONTEST)
            if prevwin:  # warn override
                self.log.info('Overwriting contest winner: ' + repr(prevwin))
            wno = u''
            wstr = u''
            lno = u''
            lstr = u''
            fstr = u''
            ft = self.contests.get_value(i, COL_200M)
            if ft is not None:
                fstr = ft.rawtime(2)
            if win == 'A':
                self.contests.set_value(i, COL_A_PLACE, wplace)
                self.contests.set_value(i, COL_B_PLACE, lplace)
                wno = self.contests.get_value(i, COL_A_NO)
                wstr = self.contests.get_value(i, COL_A_STR).decode('utf-8')
                lno = self.contests.get_value(i, COL_B_NO)
                lstr = self.contests.get_value(i, COL_B_STR).decode('utf-8')
                self.contests.set_value(i, COL_WINNER, wno)
            else:
                self.contests.set_value(i, COL_B_PLACE, wplace)
                self.contests.set_value(i, COL_A_PLACE, lplace)
                wno = self.contests.get_value(i, COL_B_NO)
                wstr = self.contests.get_value(i, COL_B_STR).decode('utf-8')
                lno = self.contests.get_value(i, COL_A_NO)
                lstr = self.contests.get_value(i, COL_A_STR).decode('utf-8')
                self.contests.set_value(i, COL_WINNER, wno)
            if not prevwin:
                self.do_places(cid, wno, wstr, wplace, lno, lstr, lplace, fstr)
                self.meet.gemini.set_bib(wno)
                self.meet.gemini.set_time(fstr.strip().rjust(4) + ' ')
                self.meet.gemini.show_brt()

    def redo_places(self):
        i = self.current_contest_combo.get_active_iter()
        if i is not None:  # contest selected ok
            cid = self.contests.get_value(i, COL_CONTEST)
            win = self.contests.get_value(i, COL_WINNER)
            wno = u''
            wstr = u''
            wplace = u''
            lno = u''
            lstr = u''
            fstr = u''
            ft = self.contests.get_value(i, COL_200M)
            if ft is not None:
                fstr = ft.rawtime(2)
            if win == 'A':
                wplace = self.contests.get_value(i, COL_A_PLACE)
                lplace = self.contests.get_value(i, COL_B_PLACE)
                wno = self.contests.get_value(i, COL_A_NO)
                wstr = self.contests.get_value(i, COL_A_STR).decode('utf-8')
                lno = self.contests.get_value(i, COL_B_NO)
                lstr = self.contests.get_value(i, COL_B_STR).decode('utf-8')
            else:
                wplace = self.contests.get_value(i, COL_B_PLACE)
                lplace = self.contests.get_value(i, COL_A_PLACE)
                wno = self.contests.get_value(i, COL_B_NO)
                wstr = self.contests.get_value(i, COL_B_STR).decode('utf-8')
                lno = self.contests.get_value(i, COL_A_NO)
                lstr = self.contests.get_value(i, COL_A_STR).decode('utf-8')
            self.do_places(cid, wno, wstr, wplace, lno, lstr, lplace, fstr)

    def do_places(self, contest, winno, winner, winpl, loseno, loser, losepl,
                  ftime):
        """Show contest result on scoreboard."""
        self.meet.scbwin = None
        self.timerwin = False
        startlist = [[u'1.', winno, winner], [u'2.', loseno, loser]]
        if ftime:
            startlist.append([u'', u'', u''])
            startlist.append([u'', u'', u'200m: ' + ftime])
        name_w = self.meet.scb.linelen - 8
        fmt = [(3, 'l'), (4, 'r'), u' ', (name_w, 'l')]
        self.meet.scbwin = scbwin.scbintsprint(
            self.meet.scb, self.meet.racenamecat(self.event), contest, fmt,
            startlist)

        self.meet.announce.gfx_overlay(2)
        self.meet.announce.gfx_set_title(u'Result: ' + self.event[u'pref'] +
                                         u' ' + self.event[u'info'] + u' ' +
                                         contest)
        self.meet.announce.gfx_add_row([winpl, winner, ftime])
        self.meet.announce.gfx_add_row([losepl, loser, ''])

        self.meet.scbwin.reset()

    def do_startlist(self):
        """Show start list on scoreboard."""
        # clear gem board
        self.meet.gemini.reset_fields()
        self.meet.gemini.show_brt()

        # prepare start list board	(use 2+2)
        cid = ''
        startlist = []
        i = self.current_contest_combo.get_active_iter()
        if i is not None:  # contest selected ok
            cid = self.contests.get_value(i, COL_CONTEST)
            self.meet.announce.gfx_overlay(2)
            self.meet.announce.gfx_set_title(u'Startlist: ' +
                                             self.event[u'pref'] + u' ' +
                                             self.event[u'info'] + u' ' + cid)
            an = self.contests.get_value(i, COL_A_NO)
            ar = self.contests.get_value(i, COL_A_STR).decode('utf-8')
            self.meet.announce.gfx_add_row([an, ar, ''])
            startlist.append([an, ar])
            bn = self.contests.get_value(i, COL_B_NO)
            br = self.contests.get_value(i, COL_B_STR).decode('utf-8')
            self.meet.announce.gfx_add_row([bn, br, ''])
            startlist.append([bn, br])
        self.meet.scbwin = None
        self.timerwin = False
        name_w = self.meet.scb.linelen - 5
        fmt = [(4, 'r'), u' ', (name_w, 'l')]
        self.meet.scbwin = scbwin.scbintsprint(
            self.meet.scb, self.meet.racenamecat(self.event), cid, fmt,
            startlist)
        self.meet.scbwin.reset()

    def update_expander_lbl_cb(self):
        """Update race info expander label."""
        self.info_expand.set_label('Race Info : ' +
                                   self.meet.racenamecat(self.event, 64))

    def editent_cb(self, entry, col):
        """Shared event entry update callback."""
        if col == u'pref':
            self.event[u'pref'] = entry.get_text().decode('utf-8', 'replace')
        elif col == u'info':
            self.event[u'info'] = entry.get_text().decode('utf-8', 'replace')
        self.update_expander_lbl_cb()

    def starttrig(self, e):
        """React to start trigger."""
        if self.timerstat == 'armstart':
            self.start = e
            self.lstart = tod.now()
            self.setrunning()
            glib.timeout_add_seconds(4, self.armfinish)

    def fintrig(self, e):
        """React to finish trigger."""
        if self.timerstat == 'armfinish':
            self.finish = e
            self.setfinished()
            self.set_elapsed()
            cid = ''
            i = self.current_contest_combo.get_active_iter()
            if i is not None:  # contest selected ok
                cid = self.contests.get_value(i, COL_CONTEST)
                self.contests.set_value(i, COL_200M, self.curelap)
                self.ctrl_winner.grab_focus()
            self.log_elapsed(cid)
            if self.timerwin and type(self.meet.scbwin) is scbwin.scbtimer:
                self.showtimer()
                if self.start is not None:
                    self.meet.gemini.rtick(self.finish - self.start, 2)
            glib.idle_add(self.delayed_announce)

    def timercb(self, e):
        """Handle a timer event."""
        chan = timy.chan2id(e.chan)
        if chan == self.startchan or chan == 0:
            self.log.debug(u'Got a start impulse.')
            self.starttrig(e)
        elif chan == self.finchan:
            self.log.debug(u'Got a finish impulse.')
            self.fintrig(e)
        return False

    def timeout(self):
        """Update scoreboard and respond to timing events."""
        if not self.winopen:
            return False
        if self.finish is None:
            self.set_elapsed()
            if self.timerwin and type(self.meet.scbwin) is scbwin.scbtimer:
                elapstr = self.time_lbl.get_text()
                self.meet.scbwin.settime(elapstr)
                self.meet.gemini.set_time(elapstr.strip().rjust(4) + ' ')
                self.meet.gemini.show_brt()
        return True

    def set_start(self, start='', lstart=None):
        """Set the race start."""
        if type(start) is tod.tod:
            self.start = start
            if lstart is not None:
                self.lstart = lstart
            else:
                self.lstart = self.start
        else:
            self.start = tod.mktod(start)
            if lstart is not None:
                self.lstart = tod.mktod(lstart)
            else:
                self.lstart = self.start
        if self.start is None:
            pass
        else:
            if self.finish is None:
                self.setrunning()

    def log_elapsed(self, contest=''):
        """Log race elapsed time on Timy."""
        if contest:
            self.meet.timer.printline(u'Ev ' + self.evno + u' [' + contest +
                                      u']')
        self.meet.timer.printline(u'      ST: ' + self.start.timestr(4))
        self.meet.timer.printline(u'     FIN: ' + self.finish.timestr(4))
        self.meet.timer.printline(u'    TIME: ' +
                                  (self.finish - self.start).timestr(2))

    def set_finish(self, finish=''):
        """Set the race finish."""
        if type(finish) is tod.tod:
            self.finish = finish
        else:
            self.finish = tod.mktod(finish)
        if self.finish is None:
            if self.start is not None:
                self.setrunning()
        else:
            if self.start is None:
                self.set_start('0')
            self.setfinished()

    def set_elapsed(self):
        """Update elapsed time in race ui and announcer."""
        self.curelap = None
        if self.start is not None and self.finish is not None:
            et = self.finish - self.start
            self.time_lbl.set_text(et.timestr(2))
            self.curelap = et
        elif self.start is not None:  # Note: uses 'local start' for RT
            runtm = (tod.now() - self.lstart).timestr(1)
            self.time_lbl.set_text(runtm)
        elif self.timerstat == 'armstart':
            self.time_lbl.set_text(tod.tod(0).timestr(1))
        else:
            self.time_lbl.set_text('')

    def current_contest_combo_changed_cb(self, combo, data=None):
        """Copy elapsed time into timer (dodgey)."""
        self.resettimer()
        i = self.current_contest_combo.get_active_iter()
        if i is not None:  # contest selected ok
            ft = self.contests.get_value(i, COL_200M)
            if ft is not None:
                self.start = tod.tod(0)
                self.finish = ft
                self.set_elapsed()
            else:
                self.start = None
                self.finish = None
                self.set_elapsed()
            winner = self.contests.get_value(i, COL_WINNER)
            self.ctrl_winner.set_text(winner)

    def race_ctrl_winner_activate_cb(self, entry, data=None):
        """Manual entry of race winner."""
        winner = entry.get_text().decode('ascii', 'ignore')
        i = self.current_contest_combo.get_active_iter()
        if i is not None:  # contest selected ok
            cid = self.contests.get_value(i, COL_CONTEST)
            self.ctrl_winner.grab_focus()
            ano = self.contests.get_value(i, COL_A_NO)
            bno = self.contests.get_value(i, COL_B_NO)
            if winner == ano:
                self.set_winner(u'A')
                glib.idle_add(self.delayed_announce)
            elif winner == bno:
                self.set_winner(u'B')
                glib.idle_add(self.delayed_announce)
            else:
                self.log.error(u'Ignored rider not in contest.')
        else:
            self.log.info(u'No contest selected.')

    def race_info_time_edit_activate_cb(self, button):
        """Display contest timing edit dialog."""
        ostx = ''
        oftx = ''
        if self.start is not None:
            ostx = self.start.rawtime(4)
        else:
            ostx = '0.0'
        if self.finish is not None:
            oftx = self.finish.rawtime(4)
        ret = uiutil.edit_times_dlg(self.meet.window, ostx, oftx)
        if ret[0] == 1:
            try:
                stod = None
                if ret[1]:
                    stod = tod.tod(ret[1], 'MANU', 'C0i')
                    self.meet.timer.printline(' ' + str(stod))
                ftod = None
                if ret[2]:
                    ftod = tod.tod(ret[2], 'MANU', 'C1i')
                    self.meet.timer.printline(' ' + str(ftod))
                self.set_start(stod)
                self.set_finish(ftod)
                self.set_elapsed()
                cid = ''
                i = self.current_contest_combo.get_active_iter()
                if i is not None:  # contest selected ok
                    cid = self.contests.get_value(i, COL_CONTEST)
                    self.contests.set_value(i, COL_200M, self.curelap)
                if self.start is not None and self.finish is not None:
                    self.log_elapsed(cid)
                self.log.info('Updated race times.')
            except Exception as v:
                self.log.error('Error updating times: ' + str(v))

            glib.idle_add(self.delayed_announce)
        else:
            self.log.info('Edit race times cancelled.')

    def delayed_announce(self):
        """Initialise the announcer's screen after a delay."""
        if self.winopen:
            self.meet.announce.clrall()
            self.meet.ann_title(' '.join([
                self.meet.event_string(self.evno), ':', self.event[u'pref'],
                self.event[u'info']
            ]))
            lapstring = strops.lapstring(self.event[u'laps'])
            substr = u' '.join(
                [lapstring, self.event[u'dist'], self.event[u'prog']]).strip()
            if substr:
                self.meet.announce.postxt(1, 0, substr.center(80))
            self.meet.announce.linefill(2, '_')
            self.meet.announce.linefill(8, '_')
            # announce current contest
            i = self.current_contest_combo.get_active_iter()
            if i is not None:  # contest selected ok
                cid = self.contests.get_value(i, COL_CONTEST)
                self.meet.announce.postxt(4, 0, u'Contest: ' + cid)
                ano = self.contests.get_value(i, COL_A_NO).rjust(3)
                astr = self.contests.get_value(i, COL_A_STR).decode('utf-8')
                aplace = self.contests.get_value(i, COL_A_PLACE).ljust(3)
                bni = self.contests.get_value(i, COL_B_NO)
                bno = bni.rjust(3)
                bstr = self.contests.get_value(i, COL_B_STR).decode('utf-8')
                bplace = self.contests.get_value(i, COL_B_PLACE).ljust(3)
                if self.contests.get_value(i, COL_WINNER) == bni:
                    self.meet.announce.postxt(
                        6, 0, bplace + u' ' + bno + u' ' + bstr)
                    self.meet.announce.postxt(
                        7, 0, aplace + u' ' + ano + u' ' + astr)
                else:
                    self.meet.announce.postxt(
                        6, 0, aplace + u' ' + ano + u' ' + astr)
                    self.meet.announce.postxt(
                        7, 0, bplace + u' ' + bno + u' ' + bstr)
                ft = self.contests.get_value(i, COL_200M)
                if ft is not None:
                    self.meet.announce.postxt(
                        6, 60, u'200m: ' + ft.rawtime(2).rjust(10))
                    self.meet.announce.postxt(
                        7, 60, u' Avg: ' + ft.speedstr().strip().rjust(10))
            # show 'leaderboard'
            lof = 10
            for c in self.contests:
                cid = (c[COL_CONTEST] + u':').ljust(8)
                win = c[COL_WINNER]
                lr = u''
                rr = u''
                sep = u' v '
                if win:
                    if c[COL_CONTEST] == u'bye':
                        sep = '   '
                    else:
                        sep = 'def'
                if win == c[COL_B_NO]:
                    lr = (c[COL_B_NO].rjust(3) + u' ' +
                          strops.truncpad(c[COL_B_STR].decode('utf-8'), 29))
                    rr = (c[COL_A_NO].rjust(3) + u' ' +
                          strops.truncpad(c[COL_A_STR].decode('utf-8'), 29))
                else:
                    lr = (c[COL_A_NO].rjust(3) + u' ' +
                          strops.truncpad(c[COL_A_STR].decode('utf-8'), 29))
                    rr = (c[COL_B_NO].rjust(3) + u' ' +
                          strops.truncpad(c[COL_B_STR].decode('utf-8'), 29))
                self.meet.announce.postxt(lof, 0, u' '.join([cid, lr, sep,
                                                             rr]))
                lof += 1

        return False

    def result_gen(self):
        """Generator function to export a final result."""
        cstack = []
        placeoft = 1
        for c in self.contests:
            rank = None
            time = None
            info = None  # rel/dsq/etc?
            win = c[COL_A_NO]
            lose = c[COL_B_NO]
            lr = False
            if c[COL_WINNER]:
                rank = placeoft
                win = c[COL_WINNER]
                if lose == win:  # win went to 'B' rider
                    lose = c[COL_A_NO]
                if c[COL_200M] is not None:
                    time = c[COL_200M].truncate(2)  # valid?
                lr = True  # include rank on loser rider
            cstack.insert(0, (lose, lr))
            time = None
            yield [win, rank, time, info]
            placeoft += 1
        for (bib, lr) in cstack:
            rank = None
            time = None
            info = None  # rel/dsq/etc?
            if lr:
                rank = placeoft
            yield [bib, rank, time, info]
            placeoft += 1

    def result_report(self, recurse=False):
        """Return a list of report sections containing the race result."""
        ret = []
        sec = report.sprintround()
        sec.heading = u'Event ' + self.evno + u': ' + u' '.join(
            [self.event[u'pref'], self.event[u'info']]).strip()
        sec.lines = []
        lapstring = strops.lapstring(self.event[u'laps'])
        substr = u' '.join(
            [lapstring, self.event[u'dist'], self.event[u'prog']]).strip()
        if substr:
            sec.subheading = substr

        for cr in self.contests:
            # if winner set, report a result, otherwise, use startlist style:
            cprompt = cr[COL_CONTEST] + u':'
            if cr[COL_WINNER]:
                avec = [
                    cr[COL_A_PLACE], cr[COL_A_NO],
                    cr[COL_A_STR].decode('utf-8'), None
                ]
                bvec = [
                    cr[COL_B_PLACE], cr[COL_B_NO],
                    cr[COL_B_STR].decode('utf-8'), None
                ]
                ft = None
                if cr[COL_200M] is not None:
                    ft = cr[COL_200M].rawtime(2)
                else:
                    ft = u' '
                if cr[COL_WINNER] == cr[COL_A_NO]:
                    sec.lines.append([cprompt, avec, bvec, ft])
                else:
                    sec.lines.append([cprompt, bvec, avec, ft])
            else:
                sec.lines.append([
                    cprompt,
                    [None, cr[COL_A_NO], cr[COL_A_STR].decode('utf-8'), None],
                    [None, cr[COL_B_NO], cr[COL_B_STR].decode('utf-8'), None],
                    None
                ])
        ret.append(sec)

        if self.comment:
            sec = report.bullet_text()
            sec.subheading = u'Decisions of the commisaires panel'
            sec.lines.append([None, self.comment])
            ret.append(sec)

        return ret

    def todstr(self, col, cr, model, iter, data=None):
        """Format tod into text for listview."""
        ft = model.get_value(iter, COL_200M)
        if ft is not None:
            cr.set_property('text', ft.rawtime(2))
        else:
            cr.set_property('text', u'')

    def destroy(self):
        """Signal race shutdown."""
        self.frame.destroy()

    def show(self):
        """Show race window."""
        self.frame.show()

    def hide(self):
        """Hide race window."""
        self.frame.hide()

    def __init__(self, meet, event, ui=True):
        """Constructor.

        Parameters:

            meet -- handle to meet object
            event -- event object handle
            ui -- display user interface?

        """
        self.meet = meet
        self.event = event
        self.evno = event[u'evid']
        self.evtype = event[u'type']
        self.series = event[u'seri']
        self.configpath = meet.event_configfile(self.evno)

        self.log = logging.getLogger('sprintround')
        self.log.setLevel(logging.DEBUG)  # config may override?
        self.log.debug(u'Creating new event: ' + repr(self.evno))

        self.readonly = not ui
        self.onestart = False
        self.start = None
        self.lstart = None
        self.finish = None
        self.curelap = None
        self.winopen = ui  # window 'open' on proper load- or consult edb
        self.timerwin = False
        self.timerstat = 'idle'
        self.autospec = ''  # automatic startlist
        self.inomnium = False
        self.startchan = 4
        self.finchan = 1
        self.contests = []
        self.comment = u''

        self.contests = gtk.ListStore(
            gobject.TYPE_STRING,  # COL_CONTEST = 0
            gobject.TYPE_STRING,  # COL_A_NO = 1
            gobject.TYPE_STRING,  # COL_A_STR = 2
            gobject.TYPE_STRING,  # COL_A_PLACE = 3
            gobject.TYPE_STRING,  # COL_B_NO = 4
            gobject.TYPE_STRING,  # COL_B_STR = 5
            gobject.TYPE_STRING,  # COL_B_PLACE = 6
            gobject.TYPE_PYOBJECT,  # COL_200M = 7
            gobject.TYPE_STRING,  # COL_WINNER = 8
            gobject.TYPE_STRING  # COL_COMMENT = 9
        )

        b = gtk.Builder()
        b.add_from_file(os.path.join(metarace.UI_PATH, 'sprnd.ui'))

        self.frame = b.get_object('race_vbox')
        self.frame.connect('destroy', self.shutdown)

        # info pane
        self.info_expand = b.get_object('info_expand')
        b.get_object('race_info_evno').set_text(self.evno)
        self.showev = b.get_object('race_info_evno_show')
        self.prefix_ent = b.get_object('race_info_prefix')
        self.prefix_ent.connect('changed', self.editent_cb, u'pref')
        self.prefix_ent.set_text(self.event[u'pref'])
        self.info_ent = b.get_object('race_info_title')
        self.info_ent.connect('changed', self.editent_cb, u'info')
        self.info_ent.set_text(self.event[u'info'])

        self.time_lbl = b.get_object('race_info_time')
        self.time_lbl.modify_font(pango.FontDescription("monospace"))
        self.type_lbl = b.get_object('race_type')
        self.type_lbl.set_text(self.event[u'type'].capitalize())

        # ctrl pane
        self.stat_but = b.get_object('race_ctrl_stat_but')
        self.ctrl_winner = b.get_object('race_ctrl_winner')
        self.ctrl_action_combo = b.get_object('race_ctrl_action_combo')
        self.ctrl_action = b.get_object('race_ctrl_action')
        self.action_model = b.get_object('race_action_model')

        self.current_contest_combo = b.get_object('current_contest_combo')
        self.current_contest_combo.set_model(self.contests)
        self.current_contest_combo.connect(
            'changed', self.current_contest_combo_changed_cb)

        # riders pane
        t = gtk.TreeView(self.contests)
        self.view = t
        t.set_reorderable(False)
        t.set_enable_search(False)
        t.set_rules_hint(True)

        # riders columns
        uiutil.mkviewcoltxt(t, 'Contest', COL_CONTEST)
        uiutil.mkviewcoltxt(t, '', COL_A_NO, calign=1.0)
        uiutil.mkviewcoltxt(t, 'A Rider', COL_A_STR, expand=True)
        uiutil.mkviewcoltxt(t, '', COL_B_NO, calign=1.0)
        uiutil.mkviewcoltxt(t, 'B Rider', COL_B_STR, expand=True)
        uiutil.mkviewcoltod(t, '200m', cb=self.todstr)
        uiutil.mkviewcoltxt(t, 'Win', COL_WINNER)
        t.show()
        b.get_object('race_result_win').add(t)

        # start timer and show window
        if ui:
            # connect signal handlers
            b.connect_signals(self)
